"""
The Element class allows parsing the LHS of a model equation.
Depending on the LHS value, either a SubscriptRange object or a Component
object will be returned. There are four components types:

- Component: Regular component, defined with '='.
- UnchangeableConstant: Unchangeable constant, defined with '=='.
- Data: Data component, defined with ':='
- Lookup: Lookup component, defined with '()'

Lookup components have their own parser for the RHS of the expression,
while the other 3 components share the same parser. The final result
from a parsed component can be exported to an AbstractComponent object
in order to build a model in other programming languages. Two more
element-like objects could be defined, which are only used for testing:

- Constraint: constraint for Reality check, defined with ':THE CONDITION:'
- TestInput: inputs for testing, defined with ':TEST INPUT:'

"""
import re
from typing import Union, Tuple, List
import warnings

import parsimonious
import numpy as np

from ..structures.abstract_model import\
    AbstractData, AbstractLookup, AbstractComponent,\
    AbstractUnchangeableConstant, AbstractSubscriptRange,\
    AbstractConstraint, AbstractTestInput

from . import vensim_utils as vu
from .vensim_structures import structures, parsing_ops


class Element():
    """
    Element object allows parsing the LHS of the Vensim expressions.

    Parameters
    ----------
    equation: str
        Original equation in the Vensim file.

    units: str
        The units of the element with the limits, i.e., the content after
        the first '~' symbol.

    documentation: str
        The comment of the element, i.e., the content after the second
        '~' symbol.

    """
    def __init__(self, equation: str, units: str, documentation: str):
        self.equation = equation
        self.units, self.limits = self._parse_units(units)
        self.documentation = documentation

    def __str__(self):  # pragma: no cover
        return "Model element:\n\t%s\nunits: %s\ndocs: %s\n" % (
            self.equation, self.units, self.documentation)

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get element information."""
        return self.__str__()

    @property
    def verbose(self):  # pragma: no cover
        """Print element information to standard output."""
        print(self._verbose)

    def _parse_units(self, units_str: str) -> Tuple[str, tuple]:
        """Separate the limits from the units."""
        # TODO improve units parsing: parse them when parsing the section
        # elements
        if not units_str:
            return "", None

        if units_str.endswith("]"):
            units, lims = units_str.rsplit("[")  # types: str, str
        else:
            return units_str, None

        lims = tuple(
            [
                float(x) if x.strip() != "?" else None
                for x in lims.strip("]").split(",")
            ]
        )
        return units.strip(), lims

    def parse(self) -> object:
        """
        Parse an Element object with parsimonious using the grammar
        given in 'parsing_grammars/element_object.peg' and the class
        ElementsComponentVisitor to visit the parsed expressions.

        Splits the LHS from the RHS of the equation. If the returned
        object is a SubscriptRange, no more parsing is needed.
        Otherwise, the RHS of the returned object (Component) should
        be parsed to get the AbstractSyntax Tree.

        Returns
        -------
        self.component: SubscriptRange or Component
            The subscript range definition object or component object.

        """
        tree = vu.Grammar.get("element_object").parse(self.equation)
        self.component = ElementsComponentVisitor(tree).component
        self.component.units = self.units
        self.component.limits = self.limits
        self.component.documentation = self.documentation
        return self.component


class ElementsComponentVisitor(parsimonious.NodeVisitor):
    """Visit model element definition to get the component object."""

    def __init__(self, ast):
        self.mapping = []
        self.subscripts = []
        self.subscripts_except = []
        self.subscripts_except_groups = []
        self.qargs = []
        self.name = None
        self.expression = None
        self.keyword = None
        self.visit(ast)

    def visit_subscript_definition(self, n, vc):
        self.component = SubscriptRange(
            self.name, self.subscripts, self.mapping)

    def visit_lookup_definition(self, n, vc):
        self.component = Lookup(
            self.name,
            (self.subscripts, self.subscripts_except_groups),
            self.expression
        )

    def visit_unchangeable_constant(self, n, vc):
        self.component = UnchangeableConstant(
            self.name,
            (self.subscripts, self.subscripts_except_groups),
            self.expression
        )

    def visit_component(self, n, vc):
        self.component = Component(
            self.name,
            (self.subscripts, self.subscripts_except_groups),
            self.expression
        )

    def visit_data_definition(self, n, vc):
        self.component = Data(
            self.name,
            (self.subscripts, self.subscripts_except_groups),
            self.keyword,
            self.expression
        )

    def visit_keyword(self, n, vc):
        self.keyword = n.text.strip()[1:-1].lower().replace(" ", "_")

    def visit_imported_subscript(self, n, vc):
        self.subscripts = dict(
            file=self.qargs[0],
            tab=self.qargs[1],
            firstcell=self.qargs[2],
            lastcell=self.qargs[3],
            prefix=self.qargs[4]
        )

    def visit_string(self, n, vc):
        self.qargs.append(vc[1])
        return vc[1]

    def visit_subscript_copy(self, n, vc):
        self.component = SubscriptRange(self.name, vc[4].strip())

    def visit_subscript_mapping(self, n, vc):
        if ":" in str(vc):
            # TODO: ensure the correct working of this condition adding
            # full integration tests
            warnings.warn(
                "\nSubscript mapping detected. "
                + "This feature works only for simple cases."
            )
            # Obtain subscript name and split by : and (
            self.mapping.append(str(vc).split(":")[0].split("(")[1].strip())
        else:
            self.mapping.append(vc[0].strip())

    def visit_subscript_range(self, n, vc):
        subs_start = re.findall(r"\d+|\D+", vc[2].strip())
        subs_end = re.findall(r"\d+|\D+", vc[6].strip())
        prefix_start, num_start = "".join(subs_start[:-1]), int(subs_start[-1])
        prefix_end, num_end = "".join(subs_end[:-1]), int(subs_end[-1])

        if not prefix_start or not prefix_end:
            raise ValueError(
                "\nA numeric range must contain at least one letter.")
        elif num_start >= num_end:
            raise ValueError(
                "\nThe number of the first subscript value must be "
                "lower than the second subscript value in a "
                "subscript numeric range.")
        elif prefix_start != prefix_end:
            raise ValueError(
                "\nOnly matching names ending in numbers are valid.")

        self.subscripts += [
            prefix_start + str(i) for i in range(num_start, num_end + 1)
            ]

    def visit_constraint_definition(self, n, vc):
        self.component = Constraint(self.name,
                                    self.subscripts,
                                    self.expression)

    def visit_test_inputs_definition(self, n, vc):
        self.component = TestInput(self.name,
                                   self.subscripts,
                                   self.expression)

    def visit_name(self, n, vc):
        self.name = vc[0].strip()

    def visit_subscript(self, n, vc):
        self.subscripts.append(n.text.strip())

    def visit_subscript_except(self, n, vc):
        self.subscripts_except.append(n.text.strip())

    def visit_subscript_except_group(self, n, vc):
        self.subscripts_except_groups.append(self.subscripts_except.copy())
        self.subscripts_except = []

    def visit_expression(self, n, vc):
        self.expression = n.text.strip()

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text


class SubscriptRange():
    """
    Subscript range definition, defined by ":" or "<->" in Vensim.
    """

    def __init__(self, name: str, definition: Union[List[str], str, dict],
                 mapping: List[str] = []):
        self.name = name
        self.definition = definition
        self.mapping = mapping

    def __str__(self):  # pragma: no cover
        return "\nSubscript range definition:  %s\n\t%s\n" % (
            self.name,
            "%s <- %s" % (self.definition, self.mapping)
            if self.mapping else self.definition)

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get subscript range information."""
        return self.__str__()

    @property
    def verbose(self):  # pragma: no cover
        """Print subscript range information to standard output."""
        print(self._verbose)

    def get_abstract_subscript_range(self) -> AbstractSubscriptRange:
        """
        Instantiates an AbstractSubscriptRange object used for building.
        This method is automatically called by the Sections's
        get_abstract_section method.

        Returns
        -------
        AbstractSubscriptRange: AbstractSubscriptRange
          AbstractSubscriptRange object that can be used for building
          the model in another programming language.

        """
        return AbstractSubscriptRange(
            name=self.name,
            subscripts=self.definition,
            mapping=self.mapping
        )


class GenericComponent():
    """
    Class to define common methods for Components, Constraints and TestInputs.
    """
    def __init__(self, name: str, subscripts: Tuple[list, list],
                 expression: str):
        self.name = name
        self.subscripts = subscripts
        self.expression = expression
        self.lookup = False
        self.ast = None
        self._kind = None

    def __str__(self):  # pragma: no cover
        text = "\n%s definition: %s" % (self._kind, self.name)
        text += "\nSubscrips: %s" % repr(self.subscripts[0])\
            if self.subscripts[0] else ""
        text += "  EXCEPT  %s" % repr(self.subscripts[1])\
            if self.subscripts[1] else ""
        text += "\n\t%s" % self._expression
        return text

    @property
    def _expression(self):  # pragma: no cover
        if hasattr(self, "ast"):
            return str(self.ast).replace("\n", "\n\t")
        else:
            return self.expression.replace("\n", "\n\t")

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get component information."""
        return self.__str__()

    @property
    def verbose(self):  # pragma: no cover
        """Print component information to standard output."""
        print(self._verbose)


class Component(GenericComponent):
    """
    Model component defined by "name = expr" in Vensim.

    Parameters
    ----------
    name: str
        The original name of the component.

    subscripts: tuple
        Tuple of length two with the list of subscripts
        in the variable definition as first argument and the list of
        subscripts that appears after the :EXCEPT: keyword (if used) as
        the second argument.

    expression: str
        The RHS of the element, expression to parse.

    """
    _kind = "Model component"

    def __init__(self, name: str, subscripts: Tuple[list, list],
                 expression: str):
        super().__init__(name, subscripts, expression)

    def parse(self) -> None:
        """
        Parse Component object with parsimonious using the grammar given
        in 'parsing_grammars/components.peg' and the class EquationVisitor
        to visit the RHS of the expressions.

        """
        tree = vu.Grammar.get("components", parsing_ops).parse(self.expression)
        self.ast = EquationVisitor(tree).translation

        if isinstance(self.ast, structures["get_xls_lookups"]):
            self.lookup = True

    def get_abstract_component(self) -> Union[AbstractComponent,
                                              AbstractLookup]:
        """
        Get Abstract Component used for building. This method is
        automatically called by Sections's get_abstract_section method.

        Returns
        -------
        AbstractComponent: AbstractComponent or AbstractLookup
          Abstract Component object that can be used for building
          the model in another language. If the component equations
          include external lookups (GET XLS/DIRECT LOOKUPS), an
          AbstractLookup class will be used.

        """
        if self.lookup:
            # get lookups equations
            return AbstractLookup(subscripts=self.subscripts, ast=self.ast)
        else:
            return AbstractComponent(subscripts=self.subscripts, ast=self.ast)


class UnchangeableConstant(Component):
    """
    Unchangeable constant defined by "name == expr" in Vensim.
    This class inherits from the Component class.

    Parameters
    ----------
    name: str
        The original name of the component.

    subscripts: tuple
        Tuple of length two with the list of subscripts
        in the variable definition as first argument and the list of
        subscripts that appears after the :EXCEPT: keyword (if used) as
        second argument.

    expression: str
        The RHS of the element, expression to parse.

    """
    _kind = "Unchangeable constant component"

    def __init__(self, name: str, subscripts: Tuple[list, list],
                 expression: str):
        super().__init__(name, subscripts, expression)

    def get_abstract_component(self) -> AbstractUnchangeableConstant:
        """
        Get Abstract Component used for building. This method is
        automatically called by Sections's get_abstract_section method.

        Returns
        -------
        AbstractComponent: AbstractUnchangeableConstant
          Abstract Component object that can be used for building
          the model in another language.

        """
        return AbstractUnchangeableConstant(
            subscripts=self.subscripts, ast=self.ast)


class Lookup(Component):
    """
    Lookup component, defined by "name(expr)" in Vensim.
    This class inherits from the Component class.

    Parameters
    ----------
    name: str
        The original name of the component.

    subscripts: tuple
        Tuple of length two with the list of subscripts in the variable
        definition as first argument and the list of subscripts that appear
        after the :EXCEPT: keyword (if used) as second argument.

    expression: str
        The RHS of the element, expression to parse.

    """
    _kind = "Lookup component"

    def __init__(self, name: str, subscripts: Tuple[list, list],
                 expression: str):
        super().__init__(name, subscripts, expression)

    def parse(self) -> None:
        """
        Parse component object with parsimonious using the grammar given
        in 'parsing_grammars/lookups.peg' and the class LookupsVisitor
        to visit the RHS of the expressions.
        """
        tree = vu.Grammar.get("lookups").parse(self.expression)
        self.ast = LookupsVisitor(tree).translation

    def get_abstract_component(self) -> AbstractLookup:
        """
        Get Abstract Component used for building. This method is
        automatically called by Sections's get_abstract_section method.

        Returns
        -------
        AbstractComponent: AbstractLookup
          Abstract Component object that may be used for building
          the model in another language.

        """
        return AbstractLookup(subscripts=self.subscripts, ast=self.ast)


class Data(Component):
    """
    Data component, defined by "name := expr" in Vensim.
    This class inherits from the Component class.

    Parameters
    ----------
    name: str
        The original name of the component.

    subscripts: tuple
        Tuple of length two with the list of subscripts in the variable
        definition as first argument and the list of subscripts that appear
        after the :EXCEPT: keyword (if used) as second argument.

    keyword: str
        The keyword used before the ":=" symbol. The following values are
        possible:  'interpolate', 'raw', 'hold_backward' and 'look_forward'.

    expression: str
        The RHS of the element, expression to parse.

    """
    _kind = "Data component"

    def __init__(self, name: str, subscripts: Tuple[list, list],
                 keyword: str, expression: str):
        super().__init__(name, subscripts, expression)
        self.keyword = keyword

    def __str__(self):  # pragma: no cover
        text = "\n%s definition: %s" % (self._kind, self.name)
        text += "\nSubscrips: %s" % repr(self.subscripts[0])\
            if self.subscripts[0] else ""
        text += "  EXCEPT  %s" % repr(self.subscripts[1])\
            if self.subscripts[1] else ""
        text += "\nKeyword: %s" % self.keyword if self.keyword else ""
        text += "\n\t%s" % self._expression
        return text

    def parse(self) -> None:
        """
        Parse component object with parsimonious using the grammar given
        in 'parsing_grammars/components.peg' and the class EquationVisitor
        to visit the RHS of the expressions.

        If the expression is None, the data will be read from a VDF file in
        Vensim.

        """
        if not self.expression:
            # empty data vars, read from vdf file
            self.ast = structures["data"]()
        else:
            super().parse()

    def get_abstract_component(self) -> AbstractData:
        """
        Get Abstract Component used for building. This method is
        automatically called by Sections's get_abstract_section method.

        Returns
        -------
        AbstractComponent: AbstractData
          Abstract Component object that can be used for building
          the model in another language.

        """
        return AbstractData(
            subscripts=self.subscripts, ast=self.ast, keyword=self.keyword)


class Constraint(GenericComponent):
    """
    Constraint definition, defined by :THE CONDITION: in Vensim.
    """

    def __init__(self, name: str, subscripts: Tuple[list, list],
                 expression: str):
        super().__init__(name, subscripts, expression)

    def __str__(self):  # pragma: no cover
        return "\nConstraint definition:  %s\n\t%s\n\t%s\n" % (
            self.name,
            self.subscripts,
            "%s" % (self.expression)
            )

    def parse(self):
        # It doesn't really parse anything, it assigns the matched expression
        # to the ast attribute
        warnings.warn("':CONSTRAINT:' detected. The expression content "
                      "is not parsed and will be ignored.")
        self.ast = self.expression

    def get_abstract_component(self) -> AbstractConstraint:
        """
        Get Abstract Component used for building. This method is
        automatically called by Sections's get_abstract_section method.

        Returns
        -------
        AbstractComponent: AbstractConstraint
          Abstract Component object that can be used for building
          the model in another language.

        """
        return AbstractConstraint(name=self.name, subscripts=self.subscripts,
                                  expression=self.ast)


class TestInput(GenericComponent):
    """
    Test Inputs definition, defined by :TEST INPUT: in Vensim.
    """

    def __init__(self, name: str, subscripts: Tuple[list, list],
                 expression: str):
        super().__init__(name, subscripts, expression)

    def __str__(self):  # pragma: no cover
        return "\nTest Inputs definition:  %s\n\t%s\n\t%s\n" % (
            self.name,
            self.subscripts,
            "%s" % (self.expression)
            )

    def parse(self):
        # It doesn't really parse anything, it assigns the matched expression
        # to the ast attribute
        warnings.warn("':TEST INPUT:' detected. The expression content "
                      "is not parsed and will be ignored.")
        self.ast = self.expression

    def get_abstract_component(self) -> AbstractTestInput:
        """
        Get Abstract Component used for building. This method is
        automatically called by Sections's get_abstract_section method.

        Returns
        -------
        AbstractComponent: AbstractTestInput
          Abstract Component object that can be used for building
          the model in another language.

        """
        return AbstractTestInput(name=self.name, subscripts=self.subscripts,
                                 expression=self.ast)


class LookupsVisitor(parsimonious.NodeVisitor):
    """Visit the elements of a lookups to get the AST"""
    def __init__(self, ast):
        self.translation = None
        self.qargs = []
        self.visit(ast)

    def visit_limits(self, n, vc):
        return n.text.strip()[:-1].replace(")-(", "),(")

    def visit_regularLookup(self, n, vc):
        if vc[0]:
            xy_limits = np.array(eval(vc[0]))
        else:
            xy_limits = np.full((2, 2), np.nan)

        values = np.array((eval(vc[2])))
        values = values[np.argsort(values[:, 0])]

        self.translation = structures["lookup"](
            x=tuple(values[:, 0]),
            y=tuple(values[:, 1]),
            x_limits=tuple(xy_limits[:, 0]),
            y_limits=tuple(xy_limits[:, 1]),
            type="interpolate"
        )

    def visit_excelLookup(self, n, vc):
        self.translation = structures["get_xls_lookups"](
            file=self.qargs[0],
            tab=self.qargs[1],
            x_row_or_col=self.qargs[2],
            cell=self.qargs[3]
        )

    def visit_string(self, n, vc):
        self.qargs.append(vc[1])
        return vc[1]

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text


class EquationVisitor(parsimonious.NodeVisitor):
    """Visit the elements of a equation to get the AST"""
    def __init__(self, ast):
        self.translation = None
        self.elements = {}
        self.subs = None  # the subscripts if given
        self.negatives = set()
        self.visit(ast)

    def visit_expr_type(self, n, vc):
        self.translation = self.elements[vc[0]]

    def visit_final_expr(self, n, vc):
        # expressions with logical binary operators (:AND:, :OR:)
        return vu.split_arithmetic(
            structures["logic"], parsing_ops["logic_ops"],
            "".join(vc).strip(), self.elements)

    def visit_logic_expr(self, n, vc):
        # expressions with logical unitary operators (:NOT:)
        id = vc[2]
        if vc[0].lower() == ":not:":
            id = self.add_element(structures["logic"](
                [":NOT:"],
                (self.elements[id],)
                ))
        return id

    def visit_comp_expr(self, n, vc):
        # expressions with comparisons (=, <>, <, <=, >, >=)
        return vu.split_arithmetic(
            structures["logic"], parsing_ops["comp_ops"],
            "".join(vc).strip(), self.elements)

    def visit_add_expr(self, n, vc):
        # expressions with additions (+, -)
        return vu.split_arithmetic(
            structures["arithmetic"], parsing_ops["add_ops"],
            "".join(vc).strip(), self.elements)

    def visit_prod_expr(self, n, vc):
        # expressions with products (*, /)
        return vu.split_arithmetic(
            structures["arithmetic"], parsing_ops["prod_ops"],
            "".join(vc).strip(), self.elements)

    def visit_exp_expr(self, n, vc):
        # expressions with exponentials (^)
        return vu.split_arithmetic(
            structures["arithmetic"], parsing_ops["exp_ops"],
            "".join(vc).strip(), self.elements, self.negatives)

    def visit_neg_expr(self, n, vc):
        id = vc[2]
        if vc[0] == "-":
            if isinstance(self.elements[id], (float, int)):
                self.elements[id] = -self.elements[id]
            else:
                self.negatives.add(id)
        return id

    def visit_call(self, n, vc):
        func = self.elements[vc[0]]
        args = self.elements[vc[4]]
        if func.reference in structures:
            return self.add_element(structures[func.reference](*args))
        else:
            return self.add_element(structures["call"](func, args))

    def visit_reference(self, n, vc):
        id = self.add_element(structures["reference"](
            vc[0].lower().replace(" ", "_"), self.subs))
        self.subs = None
        return id

    def visit_limits(self, n, vc):
        return self.add_element(n.text.strip()[:-1].replace(")-(", "),("))

    def visit_lookup_with_def(self, n, vc):
        if vc[10]:
            xy_limits = np.array(eval(self.elements[vc[10]]))
        else:
            xy_limits = np.full((2, 2), np.nan)

        values = np.array((eval(vc[11])))
        values = values[np.argsort(values[:, 0])]

        lookup = structures["lookup"](
            x=tuple(values[:, 0]),
            y=tuple(values[:, 1]),
            x_limits=tuple(xy_limits[:, 0]),
            y_limits=tuple(xy_limits[:, 1]),
            type="interpolate"
        )

        return self.add_element(structures["with_lookup"](
            self.elements[vc[4]], lookup))

    def visit_array(self, n, vc):
        if ";" in n.text or "," in n.text:
            return self.add_element(np.squeeze(np.array(
                [row.split(",") for row in n.text.strip(";").split(";")],
                dtype=float)))
        else:
            return self.add_element(eval(n.text))

    def visit_tabbed_array_call(self, n, vc):
        return self.add_element(np.array(vc[4], dtype=float))

    def visit_array_tabbed(self, n, vc):
        return n.text.strip().split()

    def visit_subscript_list(self, n, vc):
        subs = [x.strip() for x in vc[2].split(",")]
        self.subs = structures["subscripts_ref"](subs)
        return ""

    def visit_name(self, n, vc):
        return n.text.strip()

    def visit_expr(self, n, vc):
        if vc[0] not in self.elements:
            return self.add_element(eval(vc[0]))
        else:
            return vc[0]

    def visit_string(self, n, vc):
        return self.add_element(eval(n.text))

    def visit_arguments(self, n, vc):
        arglist = tuple(x.strip(",") for x in vc)
        return self.add_element(tuple(
            self.elements[arg] if arg in self.elements
            else eval(arg) for arg in arglist))

    def visit_parens(self, n, vc):
        return vc[2]

    def visit__(self, n, vc):
        # handles whitespace characters
        return ""

    def visit_nan(self, n, vc):
        return self.add_element(np.nan)

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text

    def add_element(self, element):
        return vu.add_element(self.elements, element)



"""
The Element class child classes alow parsing the expressions of a
given model element. There are four tipes of elements:

- Auxiliars (Aux class): Auxiliary elements, defined with <aux>.
- Flows (Flow class): Flow elements, defined with <flow>.
- Gfs (Gf class): Lookup elements, defined with <gf>.
- Stocks (Stock class): Data component, defined with <stock>

Moreover, a fith type element is defined ControlElement, which allows parsing
the values of the model control variables (time step, initialtime, final time).

The final result from a parsed element can be exported to an
AbstractElement object in order to build a model in other language.
"""
import re
from typing import Tuple, Union, List
from lxml import etree
import parsimonious
import numpy as np

from ..structures.abstract_model import\
    AbstractElement, AbstractControlElement,\
    AbstractLookup, AbstractComponent, AbstractSubscriptRange

from ..structures.abstract_expressions import\
    AbstractSyntax, CallStructure, ReferenceStructure

from . import xmile_utils as vu
from .xmile_structures import structures, parsing_ops


class Element():
    """
    Element class. This class provides the shared methods for its
    children: Aux, Flow, Gf, Stock, and ControlElement.

    Parameters
    ----------
    node: etree._Element
        The element node content.

    ns: dict
        The namespace of the section.

    subscripts: dict
        The subscript dictionary of the section, necessary to parse
        some subscripted elements.

    """
    _interp_methods = {
        "continuous": "interpolate",
        "extrapolate": "extrapolate",
        "discrete": "hold_backward"
    }

    _kind = "Element"

    def __init__(self, node: etree._Element, ns: dict, subscripts):
        self.node = node
        self.ns = ns
        self.name = node.attrib["name"].replace("\\n", " ")
        self.units = self._get_xpath_text(node, "ns:units") or ""
        self.documentation = self._get_xpath_text(node, "ns:doc") or ""
        self.limits = (None, None)
        self.components = []
        self.subscripts = subscripts

    def __str__(self):  # pragma: no cover
        text = "\n%s definition: %s" % (self._kind, self.name)
        text += "\nSubscrips: %s" % repr(self.subscripts)\
            if self.subscripts else ""
        text += "\n\t%s" % self._expression
        return text

    @property
    def _expression(self):  # pragma: no cover
        if hasattr(self, "ast"):
            return str(self.ast).replace("\n", "\n\t")

        else:
            return self.node.text.replace("\n", "\n\t")

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get element information."""
        return self.__str__()

    @property
    def verbose(self):  # pragma: no cover
        """Print element information to standard output."""
        print(self._verbose)

    def _get_xpath_text(self, node: etree._Element,
                        xpath: str) -> Union[str, None]:
        """Safe access of occassionally missing text"""
        try:
            return node.xpath(xpath, namespaces=self.ns)[0].text
        except IndexError:
            return None

    def _get_xpath_attrib(self, node: etree._Element,
                          xpath: str, attrib: str) -> Union[str, None]:
        """Safe access of occassionally missing attributes"""
        # defined here to take advantage of NS in default
        try:
            return node.xpath(xpath, namespaces=self.ns)[0].attrib[attrib]
        except IndexError:
            return None

    def _get_limits(self) -> Tuple[Union[None, str], Union[None, str]]:
        """Get the limits of the element"""
        lims = (
            self._get_xpath_attrib(self.node, 'ns:range', 'min'),
            self._get_xpath_attrib(self.node, 'ns:range', 'max')
        )
        return tuple(float(x) if x is not None else x for x in lims)

    def _get_non_negative(self, behavior):
        non_negative = behavior or bool(
            self.node.xpath('ns:non_negative', namespaces=self.ns)
        )
        boolean = self._get_xpath_text(self.node, 'ns:non_negative')
        if boolean is not None:
            non_negative = 'false' not in boolean.lower()
        return non_negative

    def _parse_lookup_xml_node(self, node: etree._Element) -> AbstractSyntax:
        """
        Parse lookup definition

        Returns
        -------
        AST: AbstractSyntax

        """
        ys_node = node.xpath('ns:ypts', namespaces=self.ns)[0]
        ys = np.fromstring(
            ys_node.text,
            dtype=float,
            sep=ys_node.attrib['sep'] if 'sep' in ys_node.attrib else ','
        )
        xscale_node = node.xpath('ns:xscale', namespaces=self.ns)
        if len(xscale_node) > 0:
            xmin = xscale_node[0].attrib['min']
            xmax = xscale_node[0].attrib['max']
            xs = np.linspace(float(xmin), float(xmax), len(ys))
        else:
            xs_node = node.xpath('ns:xpts', namespaces=self.ns)[0]
            xs = np.fromstring(
                xs_node.text,
                dtype=float,
                sep=xs_node.attrib['sep'] if 'sep' in xs_node.attrib else ','
            )

        interp = node.attrib['type'] if 'type' in node.attrib else 'continuous'

        return structures["lookup"](
            x=tuple(xs[np.argsort(xs)]),
            y=tuple(ys[np.argsort(xs)]),
            x_limits=(np.min(xs), np.max(xs)),
            y_limits=(np.min(ys), np.max(ys)),
            type=self._interp_methods[interp]
        )

    def parse(self, behaviors: dict) -> None:
        """
        Parse all the components of an element

        Parameters
        ----------
        behaviors: dict
            Dictionary with keys 'non_negative_flow' and 'non_negative_stock'
            and boolean values defining the global behavior for the
            stocks and flows.

        Returns
        -------
        None

        """
        if self.node.xpath("ns:element", namespaces=self.ns):
            # defined in several equations each with one subscript
            for subnode in self.node.xpath("ns:element", namespaces=self.ns):
                self.components.append(
                    ((subnode.attrib["subscript"].split(","), []),
                     self._parse_component(subnode, behaviors)[0])
                )
        else:
            # get the subscripts from element
            subscripts = [
                subnode.attrib["name"]
                for subnode
                in self.node.xpath("ns:dimensions/ns:dim", namespaces=self.ns)
            ]
            parsed = self._parse_component(self.node, behaviors)
            if len(parsed) == 1:
                # element defined with one equation
                self.components = [((subscripts, []),  parsed[0])]
            else:
                # element defined in several equations, but only the general
                # subscripts are given, save each equation with its
                # subscrtipts
                subs_list = self.subscripts[subscripts[0]]
                self.components = [
                    (([subs], []), parsed_i) for subs, parsed_i in
                    zip(subs_list, parsed)
                ]

    def _smile_parser(self, expression: str) -> AbstractSyntax:
        """
        Parse expression with parsimonious.

        Returns
        -------
        AST: AbstractSyntax

        """
        tree = vu.Grammar.get("equations", parsing_ops).parse(
            expression.strip())
        return EquationVisitor(tree).translation

    def _get_empty_abstract_element(self) -> AbstractElement:
        """
        Get empty Abstract used for building

        Returns
        -------
        AbstractElement
        """
        return AbstractElement(
            name=self.name,
            units=self.units,
            limits=self.limits,
            documentation=self.documentation,
            components=[])


class Aux(Element):
    """
    Auxiliary variable defined by <aux> in Xmile.

    Parameters
    ----------
    node: etree._Element
        The element node content.

    ns: dict
        The namespace of the section.

    subscripts: dict
        The subscript dictionary of the section, necessary to parse
        some subscripted elements.

    """
    _kind = "Flow"

    def __init__(self, node, ns, subscripts):
        super().__init__(node, ns, subscripts)
        self.limits = self._get_limits()

    def _parse_component(self, node: etree._Element,
                         behaviors: dict) -> List[AbstractSyntax]:
        """
        Parse one Aux component

        Returns
        -------
        AST: AbstractSyntax

        """
        asts = []
        for eqn in node.xpath('ns:eqn', namespaces=self.ns):
            # Replace new lines with space, and replace 2 or more spaces with
            # single space. Then ensure there is no space at start or end of
            # equation
            eqn = re.sub(r"(\s{2,})", " ", eqn.text.replace("\n", ' ')).strip()
            ast = self._smile_parser(eqn)

            gf_node = self.node.xpath("ns:gf", namespaces=self.ns)
            if len(gf_node) > 0:
                ast = structures["inline_lookup"](
                    ast, self._parse_lookup_xml_node(gf_node[0]))
            asts.append(ast)

        return asts

    def get_abstract_element(self) -> AbstractElement:
        """
        Get Abstract Element used for building. This method is
        automatically called by Sections's get_abstract_section.

        Returns
        -------
        AbstractElement: AbstractElement
          Abstract Element object that can be used for building
          the model in another language. It contains a list of
          AbstractComponents with the Abstract Syntax Tree of each of
          the expressions.

        """
        ae = self._get_empty_abstract_element()
        for component in self.components:
            ae.components.append(AbstractComponent(
                subscripts=component[0],
                ast=component[1]))
        return ae


class Flow(Aux):
    """
    Flow defined by <flow> in Xmile.

    Parameters
    ----------
    node: etree._Element
        The element node content.

    ns: dict
        The namespace of the section.

    subscripts: dict
        The subscript dictionary of the section, necessary to parse
        some subscripted elements.

    """
    _kind = "Flow"

    def __init__(self, node, ns, subscripts):
        super().__init__(node, ns, subscripts)

    def _parse_component(self, node: etree._Element,
                         behaviors: dict) -> List[AbstractSyntax]:
        """
        Parse one Flow component

        Returns
        -------
        AST: AbstractSyntax

        """
        asts = super()._parse_component(node, behaviors)
        if self._get_non_negative(behaviors['non_negative_flow']):
            # non_negative flows
            asts = [
                CallStructure(
                    ReferenceStructure("max"), (ast, 0)
                )
                for ast in asts
            ]

        return asts


class Gf(Element):
    """
    Gf variable (lookup) defined by <gf> in Xmile.

    Parameters
    ----------
    node: etree._Element
        The element node content.

    ns: dict
        The namespace of the section.

    subscripts: dict
        The subscript dictionary of the section, necessary to parse
        some subscripted elements.

    """
    _kind = "Gf component"

    def __init__(self, node, ns, subscripts):
        super().__init__(node, ns, subscripts)
        self.limits = self.get_limits()

    def get_limits(self) -> Tuple[Union[None, str], Union[None, str]]:
        """Get the limits of the Gf element"""
        lims = (
            self._get_xpath_attrib(self.node, 'ns:yscale', 'min'),
            self._get_xpath_attrib(self.node, 'ns:yscale', 'max')
        )
        return tuple(float(x) if x is not None else x for x in lims)

    def _parse_component(self, node: etree._Element,
                         behaviors: dict) -> AbstractSyntax:
        """
        Parse one Gf component

        Returns
        -------
        AST: AbstractSyntax

        """
        return [self._parse_lookup_xml_node(self.node)]

    def get_abstract_element(self) -> AbstractElement:
        """
        Get Abstract Element used for building. This method is
        automatically called by Sections's get_abstract_section.

        Returns
        -------
        AbstractElement: AbstractElement
          Abstract Element object that can be used for building
          the model in another language. It contains a list of
          AbstractComponents with the Abstract Syntax Tree of each of
          the expressions.

        """
        ae = self._get_empty_abstract_element()
        for component in self.components:
            ae.components.append(AbstractLookup(
                subscripts=component[0],
                ast=component[1]))
        return ae


class Stock(Element):
    """
    Stock variable defined by <stock> in Xmile.

    Parameters
    ----------
    node: etree._Element
        The element node content.

    ns: dict
        The namespace of the section.

    subscripts: dict
        The subscript dictionary of the section, necessary to parse
        some subscripted elements.

    """

    _kind = "Stock component"

    def __init__(self, node, ns, subscripts):
        super().__init__(node, ns, subscripts)
        self.limits = self._get_limits()

    def _parse_component(self, node, behaviors: dict) -> AbstractSyntax:
        """
        Parse one Stock component

        Returns
        -------
        AST: AbstractSyntax

        """
        # Parse each flow equations
        inflows = [
            self._smile_parser(inflow.text)
            for inflow in self.node.xpath('ns:inflow', namespaces=self.ns)]
        outflows = [
            self._smile_parser(outflow.text)
            for outflow in self.node.xpath('ns:outflow', namespaces=self.ns)]

        if inflows:
            # stock has inflows
            expr = ["+"] * (len(inflows)-1) + ["-"] * len(outflows)
        elif outflows:
            # stock has no inflows but outflows
            outflows[0] = structures["negative"](outflows[0])
            expr = ["-"] * (len(outflows)-1)
        else:
            # stock is constant
            expr = []
            inflows = [0]

        if expr:
            # stock has more than one flow
            flows = structures["arithmetic"](expr, inflows+outflows)
        else:
            # stock has only one flow
            flows = inflows[0] if inflows else outflows[0]

        # Read the initial value equation for stock element
        initial = self._smile_parser(self._get_xpath_text(self.node, 'ns:eqn'))

        # Get non-negative information
        non_negative = self._get_non_negative(behaviors['non_negative_stock'])

        return [structures["stock"](flows, initial, non_negative)]

    def get_abstract_element(self) -> AbstractElement:
        """
        Get Abstract Element used for building. This method is
        automatically called by Sections's get_abstract_section.

        Returns
        -------
        AbstractElement: AbstractElement
          Abstract Element object that can be used for building
          the model in another language. It contains a list of
          AbstractComponents with the Abstract Syntax Tree of each of
          the expressions.

        """
        ae = self._get_empty_abstract_element()
        for component in self.components:
            ae.components.append(AbstractComponent(
                subscripts=component[0],
                ast=component[1]))
        return ae


class ControlElement(Element):
    """Control variable (lookup)"""
    _kind = "Control variable"

    def __init__(self, name, units, documentation, eqn):
        self.name = name
        self.units = units
        self.documentation = documentation
        self.limits = (None, None)
        self.eqn = eqn

    def parse(self, behaviors: dict) -> None:
        """
        Parse control elment.

        Parameters
        ----------
        behaviors: dict
            Dictionary with keys 'non_negative_flow' and 'non_negative_stock'
            and boolean values defining the global behavior for the
            stocks and flows.

        Returns
        -------
        None

        """
        self.ast = self._smile_parser(self.eqn)

    def get_abstract_element(self) -> AbstractElement:
        """
        Get Abstract Element used for building. This method is
        automatically called by Sections's get_abstract_section.

        Returns
        -------
        AbstractElement: AbstractElement
          Abstract Element object that can be used for building
          the model in another language. It contains an AbstractComponent
          with the Abstract Syntax Tree of the expression.

        """
        return AbstractControlElement(
            name=self.name,
            units=self.units,
            limits=self.limits,
            documentation=self.documentation,
            components=[
                AbstractComponent(subscripts=([], []), ast=self.ast)
            ]
        )


class SubscriptRange():
    """Subscript range definition."""

    def __init__(self, name: str, definition: List[str],
                 mapping: List[str] = []):
        self.name = name
        self.definition = definition
        self.mapping = mapping

    def __str__(self):  # pragma: no cover
        return "\nSubscript range definition:  %s\n\t%s\n" % (
            self.name,
            self.definition)

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get subscript range information."""
        return self.__str__()

    @property
    def verbose(self):  # pragma: no cover
        """Print subscript range information to standard output."""
        print(self._verbose)

    def get_abstract_subscript_range(self) -> AbstractSubscriptRange:
        """
        Get Abstract Subscript Range used for building. This method is
        automatically called by Sections's get_abstract_section.

        Returns
        -------
        AbstractSubscriptRange: AbstractSubscriptRange
          Abstract Subscript Range object that can be used for building
          the model in another language.

        """
        return AbstractSubscriptRange(
            name=self.name,
            subscripts=self.definition,
            mapping=self.mapping
        )


class EquationVisitor(parsimonious.NodeVisitor):
    """Visit the elements of a equation to get the AST"""
    def __init__(self, ast):
        self.translation = None
        self.elements = {}
        self.subs = None  # the subscripts if given
        self.negatives = set()
        self.visit(ast)

    def visit_expr_type(self, n, vc):
        self.translation = self.elements[vc[0]]

    def visit_logic2_expr(self, n, vc):
        # expressions with logical binary operators (and, or)
        return vu.split_arithmetic(
            structures["logic"], parsing_ops["logic_ops"],
            "".join(vc).strip(), self.elements)

    def visit_logic_expr(self, n, vc):
        # expressions with logical unitary operators (not)
        id = vc[2]
        if vc[0].lower() == "not":
            id = self.add_element(structures["logic"](
                [":NOT:"],
                (self.elements[id],)
                ))
        return id

    def visit_comp_expr(self, n, vc):
        # expressions with comparisons (=, <>, <, <=, >, >=)
        return vu.split_arithmetic(
            structures["logic"], parsing_ops["comp_ops"],
            "".join(vc).strip(), self.elements)

    def visit_add_expr(self, n, vc):
        # expressions with additions (+, -)
        return vu.split_arithmetic(
            structures["arithmetic"], parsing_ops["add_ops"],
            "".join(vc).strip(), self.elements)

    def visit_mod_expr(self, n, vc):
        # modulo expressions (mod)
        if vc[1].lower().startswith("mod"):
            return self.add_element(
                structures["call"](
                    structures["reference"]("modulo"),
                    (self.elements[vc[0]], self.elements[vc[1][3:]])
                ))
        else:
            return vc[0]

    def visit_prod_expr(self, n, vc):
        # expressions with products (*, /)
        return vu.split_arithmetic(
            structures["arithmetic"], parsing_ops["prod_ops"],
            "".join(vc).strip(), self.elements)

    def visit_exp_expr(self, n, vc):
        # expressions with exponentials (^)
        return vu.split_arithmetic(
            structures["arithmetic"], parsing_ops["exp_ops"],
            "".join(vc).strip(), self.elements, self.negatives)

    def visit_neg_expr(self, n, vc):
        id = vc[2]
        if vc[0] == "-":
            if isinstance(self.elements[id], (float, int)):
                self.elements[id] = -self.elements[id]
            else:
                self.negatives.add(id)
        return id

    def visit_call(self, n, vc):
        func = self.elements[vc[0]]
        args = self.elements[vc[4]]
        if func.reference in structures:
            func_str = structures[func.reference]
            if isinstance(func_str, dict):
                return self.add_element(func_str[len(args)](*args))
            else:
                return self.add_element(func_str(*args))
        else:
            return self.add_element(structures["call"](func, args))

    def visit_conditional_statement(self, n, vc):
        return self.add_element(structures["if_then_else"](
            self.elements[vc[2]],
            self.elements[vc[6]],
            self.elements[vc[10]]))

    def visit_reference(self, n, vc):
        id = self.add_element(structures["reference"](
            vc[0].lower().replace(" ", "_").strip("\""), self.subs))
        self.subs = None
        return id

    def visit_array(self, n, vc):
        if ";" in n.text or "," in n.text:
            return self.add_element(np.squeeze(np.array(
                [row.split(",") for row in n.text.strip(";").split(";")],
                dtype=float)))
        else:
            return self.add_element(eval(n.text))

    def visit_subscript_list(self, n, vc):
        subs = [x.strip().replace("_", " ") for x in vc[2].split(",")]
        self.subs = structures["subscripts_ref"](subs)
        return ""

    def visit_name(self, n, vc):
        return n.text.strip()

    def visit_expr(self, n, vc):
        if vc[0] not in self.elements:
            return self.add_element(eval(vc[0]))
        else:
            return vc[0]

    def visit_arguments(self, n, vc):
        arglist = tuple(x.strip(",") for x in vc)
        return self.add_element(tuple(
            self.elements[arg] if arg in self.elements
            else eval(arg) for arg in arglist))

    def visit_parens(self, n, vc):
        return vc[2]

    def visit__(self, n, vc):
        # handles whitespace characters
        return ""

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text

    def add_element(self, element):
        return vu.add_element(self.elements, element)



"""
Model components and time managing classes.
"""

from warnings import warn
import os
import random
import inspect
import importlib.util
from copy import deepcopy

import numpy as np

from pysd._version import __version__


class Component(object):

    def __init__(self):
        self.namespace = {}
        self.dependencies = {}

    def add(self, name, units=None, limits=(np.nan, np.nan),
            subscripts=None, comp_type=None, comp_subtype=None,
            depends_on={}, other_deps={}):
        """
        This decorators allows assigning metadata to a function.
        """
        def decorator(function):
            function.name = name
            function.units = units
            function.limits = limits
            function.subscripts = subscripts
            function.type = comp_type
            function.subtype = comp_subtype
            function.args = inspect.getfullargspec(function)[0]

            # include component in namespace and dependencies
            self.namespace[name] = function.__name__
            if function.__name__ != "time":
                self.dependencies[function.__name__] = depends_on
                self.dependencies.update(other_deps)

            return function

        return decorator


class Components(object):
    """
    Workaround class to let the user do:
        model.components.var = value
    """
    def __init__(self, py_model_file, set_components):
        object.__setattr__(self, "_components", self._load(py_model_file))
        object.__setattr__(self, "_set_components", set_components)

    def _load(self, py_model_file):
        """
        Load model components.

        Parameters
        ----------
        py_model_file: str
            Model file to be loaded.

        Returns
        -------
        components: module
            The imported file content.

        """
        # need a unique identifier for the imported module.
        module_name = os.path.splitext(py_model_file)[0]\
            + str(random.randint(0, 1000000))
        try:
            spec = importlib.util.spec_from_file_location(
                module_name, py_model_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        except TypeError:
            raise ImportError(
                "\n\nNot able to import the model. "
                + "This may be because the model was compiled with an "
                + "earlier version of PySD, you can check on the top of "
                + "the model file you are trying to load."
                + "\nThe current version of PySd is :"
                + "\n\tPySD " + __version__ + "\n\n"
                + "Please translate again the model with the function"
                + " read_vensim or read_xmile.")

    def __getattribute__(self, name):
        """
        Get attribute from the class. Try Except vlock is used to load directly
        model components in order to avoid making the model slower during the
        integration.
        """
        try:
            return getattr(object.__getattribute__(self, "_components"), name)
        except AttributeError:
            if name in ["_components", "_set_components",
                        "_set_component", "_load"]:
                # The attribute is from the class Components
                return object.__getattribute__(self, name)
            else:
                raise NameError(f"Component '{name}' not found in the model.")

    def __setattr__(self, name, value):
        """
        Workaround calling the Macro._set_components method
        """
        self._set_components({name: value})

    def _set_component(self, name, value):
        """
        Replaces the previous setter.
        """
        setattr(
            object.__getattribute__(self, "_components"),
            name,
            value
        )


class Time(object):
    rprec = 1e-5  # relative precision for final time and saving time

    def __init__(self):
        self._time = None
        self.stage = None
        self.return_timestamps = None
        self._next_return = None
        self._control_vars_tracker = {}

    def __call__(self):
        return self._time

    def export(self):
        """Exports time values to a dictionary."""
        return {
            "control_vars": self._get_control_vars(),
            "stage": self.stage,
            "_time": self._time,
            "return_timestamps": self.return_timestamps,
            "_next_return": self._next_return
        }

    def _get_control_vars(self):
        """
        Make control vars changes exportable.
        """
        out = {}
        for cvar, value in self._control_vars_tracker.items():
            if callable(value):
                out[cvar] = value()
            else:
                out[cvar] = value
        return out

    def _set_time(self, time_dict):
        """Copy values from other Time object, used by Model.copy"""
        self.set_control_vars(**time_dict['control_vars'])
        for key, value in time_dict.items():
            if key == 'control_vars':
                continue
            setattr(self, key, value)

    def set_control_vars(self, **kwargs):
        """
        Set the control variables values

        Parameters
        ----------
        **kwards:
            initial_time: float, callable or None
                Initial time.
            final_time: float, callable or None
                Final time.
            time_step: float, callable or None
                Time step.
            saveper: float, callable or None
                Saveper.

        """
        # filter None values
        kwargs = {
                key: value for key, value in kwargs.items()
                if value is not None
        }
        # track changes
        self._control_vars_tracker.update(kwargs)
        self._set_control_vars(**kwargs)

    def _set_control_vars(self, **kwargs):
        """
        Set the control variables values. Private version to be used
        to avoid tracking changes.
        """
        def _convert_value(value):
            # this function is necessary to avoid copying the pointer in the
            # lambda function.
            if callable(value):
                return value
            else:
                return lambda: value

        for key, value in kwargs.items():
            setattr(self, key, _convert_value(value))

        if "initial_time" in kwargs:
            self._initial_time = self.initial_time()
            self._time = self.initial_time()

    def in_bounds(self):
        """
        Check if time is smaller than current final time value.

        Returns
        -------
        bool:
            True if time is smaller than final time. Otherwise, returns Fase.

        """
        return self._time + self.time_step()*self.rprec < self.final_time()

    def in_return(self):
        """ Check if current time should be returned """
        prec = self.time_step() * self.rprec

        if self.return_timestamps is not None:
            # this allows managing float precision error
            if self._next_return is None:
                return False
            if np.isclose(self._time, self._next_return, prec):
                self._update_next_return()
                return True
            else:
                while self._next_return is not None\
                      and self._time > self._next_return:
                    warn(
                        f"The returning time stamp '{self._next_return}' "
                        "seems to not be a multiple of the time step. "
                        "This value will not be saved in the output. "
                        "Please, modify the returning timestamps or the "
                        "integration time step to avoid this."
                        )
                    self._update_next_return()
                return False

        time_delay = self._time - self._initial_time
        save_per = self.saveper()
        return time_delay % save_per < prec or -time_delay % save_per < prec

    def round(self):
        """ Return rounded time to outputs to avoid float precision error"""
        return np.round(
            self._time,
            -int(np.log10(self.time_step()*self.rprec)))

    def add_return_timestamps(self, return_timestamps):
        """ Add return timestamps """
        if hasattr(return_timestamps, '__len__')\
           and len(return_timestamps) > 0:
            self.return_timestamps = list(return_timestamps)
            self.return_timestamps.sort(reverse=True)
            self._next_return = self.return_timestamps.pop()
        elif isinstance(return_timestamps, (float, int)):
            self._next_return = return_timestamps
            self.return_timestamps = []
        else:
            self._next_return = None
            self.return_timestamps = None

    def update(self, value):
        """ Update current time value """
        self._time = value

    def _update_next_return(self):
        """ Update the next_return value """
        if self.return_timestamps:
            self._next_return = self.return_timestamps.pop()
        else:
            self._next_return = None

    def reset(self):
        """ Reset time value to the initial """
        self._time = self._initial_time



"""
These are general utilities used by the builder.py, functions.py or the
model file. Vensim's function equivalents should not go here but in
functions.py
"""

import json
from datetime import datetime
from pathlib import Path
from chardet.universaldetector import UniversalDetector
from dataclasses import dataclass
from typing import Dict, Set

import progressbar
import numpy as np
import xarray as xr
import pandas as pd


def xrsplit(array):
    """
    Split an array to a list of all the components.

    Parameters
    ----------
    array: xarray.DataArray
        Array to split.

    Returns
    -------
    sp_list: list of xarray.DataArrays
        List of shape 0 xarray.DataArrays with coordinates.

    """
    sp_list = [sa for sa in array]
    if sp_list[0].shape:
        sp_list = [ssa for sa in sp_list for ssa in xrsplit(sa)]
    return sp_list


def get_current_computer_time():
    """
    Returns the current machine time. Needed to mock the machine time in
    the tests.

    Parameters
    ---------
    None

    Returns
    -------
    datetime.now(): datetime.datetime
        Current machine time.

    """
    return datetime.now()


def get_return_elements(return_columns, namespace):
    """
    Takes a list of return elements formatted in vensim's format
    Varname[Sub1, SUb2]
    and returns first the model elements (in Python safe language)
    that need to be computed and collected, and secondly the addresses
    that each element in the return columns list translates to

    Parameters
    ----------
    return_columns: list of strings

    namespace: dict

    Returns
    -------
    capture_elements
    return_addresses

    """
    capture_elements = list()
    return_addresses = dict()
    for col in return_columns:
        if col[0] == col[-1] and col[0] == '"':
            name = col
            address = None
        elif "[" in col:
            name, location = col.strip("]").split("[")
            address = tuple([loc.strip() for loc in location.split(",")])
        else:
            name = col
            address = None

        if name in namespace:
            py_name = namespace[name]
        else:
            if name in namespace.values():
                py_name = name
            else:
                raise KeyError(name + " not found as model element")

        if py_name not in capture_elements:
            capture_elements += [py_name]

        return_addresses[col] = (py_name, address)

    return list(capture_elements), return_addresses


def compute_shape(coords, reshape_len=None, py_name=""):
    """
    Computes the 'shape' of a coords dictionary.
    Function used to rearange data in xarrays and
    to compute the number of rows/columns to be read in a file.

    Parameters
    ----------
    coords: dict
      Ordered dictionary of the dimension names as a keys with their values.

    reshape_len: int (optional)
      Number of dimensions of the output shape.
      The shape will ony compute the corresponent table
      dimensions to read from Excel, then, the dimensions
      with length one will be ignored at first.
      Lately, it will complete with 1 on the left of the shape
      if the reshape_len value is bigger than the length of shape.
      Will raise a ValueError if we try to reshape to a reshape_len
      smaller than the initial shape.

    py_name: str
      Name to print if an error is raised.

    Returns
    -------
    shape: list
      Shape of the ordered dictionary or of the desired table or vector.

    """
    if not reshape_len:
        return [len(coord) for coord in coords.values()]

    # get the shape of the coordinates bigger than 1
    shape = [len(coord) for coord in coords.values() if len(coord) > 1]

    shape_len = len(shape)

    # return an error when the current shape is bigger than the requested one
    if shape_len > reshape_len:
        raise ValueError(
            py_name
            + "\n"
            + "The shape of the coords to read in a "
            + " external file must be at most "
            + "{} dimensional".format(reshape_len)
        )

    # complete with 1s on the left
    return [1] * (reshape_len - shape_len) + shape


def get_key_and_value_by_insensitive_key_or_value(key, dict):
    """
    Providing a key or value in a dictionary search for the real key and value
    in the dictionary ignoring case sensitivity.

    Parameters
    ----------
    key: str
        Key or value to look for in the dictionary.
    dict: dict
        Dictionary to search in.

    Returns
    -------
    real key, real value: (str, str) or (None, None)
        The real key and value that appear in the dictionary or a tuple
        of Nones if the input key is not in the dictionary.

    """
    lower_key = key.lower()
    for real_key, real_value in dict.items():
        if real_key.lower() == lower_key or real_value.lower() == lower_key:
            return real_key, real_value

    return None, None


def rearrange(data, dims, coords):
    """
    Returns a xarray.DataArray object with the given coords and dims

    Parameters
    ---------
    data: float or xarray.DataArray
        The input data to rearrange.

    dims: list
        Ordered list of the dimensions.

    coords: dict
        Dictionary of the dimension names as a keys with their values.

    Returns
    -------
    xarray.DataArray

    """
    # subset used coords in general coords will be the subscript_dict
    coords = {dim: coords[dim] for dim in dims}
    if isinstance(data, xr.DataArray):
        shape = tuple(compute_shape(coords))
        if data.shape == shape:
            # Allows switching dimensions names and transpositions
            return xr.DataArray(data=data.values, coords=coords, dims=dims)
        elif np.prod(shape) < np.prod(data.shape):
            # Allows subscripting a subrange
            return data.rename({
                dim: new_dim for dim, new_dim in zip(data.dims, dims)
                if dim != new_dim
            }).loc[coords]

        # The coordinates are expanded or transposed
        return xr.DataArray(0, coords, dims) + data

    elif data is not None:
        return xr.DataArray(data, coords, dims)

    return None


def load_model_data(root, model_name):

    """
    Used for models split in several files.
    Loads subscripts and modules dictionaries

    Parameters
    ----------
    root: pathlib.Path
        Path to the model file.

    model_name: str
        Name of the model without file type extension (e.g. "my_model").

    Returns
    -------
    subscripts: dict
        Dictionary describing the possible dimensions of the stock's
        subscripts.

    modules: dict
        Dictionary containing view (module) names as keys and a list of the
        corresponding variables as values.

    """
    with open(root.joinpath("_subscripts_" + model_name + ".json")) as subs:
        subscripts = json.load(subs)

    # the _modules.json in the sketch_var folder shows to which module each
    # variable belongs
    with open(root.joinpath("modules_" + model_name, "_modules.json")) as mods:
        modules = json.load(mods)

    return subscripts, modules


def load_modules(module_name, module_content, work_dir, submodules):
    """
    Used to load model modules from the main model file, when
    split_views=True in the read_vensim function. This function is used
    to iterate over the different layers of the nested dictionary that
    describes which model variables belong to each module/submodule.

    Parameters
    ----------
    module_name: str
        Name of the module to load.

    module_content: dict or list
        Content of the module. If it's a dictionary, it means that the
        module has submodules, whereas if it is a list it means that that
        particular module/submodule is a final one.

    work_dir: pathlib.Path
        Path to the module file.

    submodules: list
        This list gets updated at every recursive iteration, and each element
        corresponds to the string representation of each module/submodule that
        is read.

    Returns
    -------
    str:
        String representations of the modules/submodules to execute in the main
        model file.

    """
    if isinstance(module_content, list):
        with open(work_dir.joinpath(module_name + ".py"), "r",
                  encoding="UTF-8") as mod:
            submodules.append(mod.read())
    else:
        for submod_name, submod_content in module_content.items():
            load_modules(
                submod_name, submod_content,
                work_dir.joinpath(module_name),
                submodules)

    return "\n\n".join(submodules)


def load_outputs(file_name, transpose=False, columns=None, encoding=None):
    """
    Load outputs file

    Parameters
    ----------
    file_name: str or pathlib.Path
        Output file to read. Must be csv or tab.

    transpose: bool (optional)
        If True reads transposed outputs file, i.e. one variable per row.
        Default is False.

    columns: list or None (optional)
        List of the column names to load. If None loads all the columns.
        Default is None.
        NOTE: if transpose=False, the loading will be faster as only
        selected columns will be loaded. If transpose=True the whole
        file must be read and it will be subselected later.

    encoding: str or None (optional)
        Encoding type to read output file. Needed if the file has special
        characters. Default is None.

    Returns
    -------
    pandas.DataFrame
        A pandas.DataFrame with the outputs values.

    """
    read_func = {'.csv': pd.read_csv, '.tab': pd.read_table}

    file_name = Path(file_name)

    if columns:
        columns = set(columns)
        if not transpose:
            columns.add("Time")

    for end, func in read_func.items():
        if file_name.suffix.lower() == end:
            if transpose:
                out = func(file_name,
                           encoding=encoding,
                           index_col=0).T
                if columns:
                    out = out[list(columns)]
            else:
                out = func(file_name,
                           encoding=encoding,
                           usecols=columns,
                           index_col="Time")

            out.index = out.index.astype(float)
            # return the dataframe removing nan index values
            return out[~np.isnan(out.index)]

    raise ValueError(
        f"\nNot able to read '{file_name}'. "
        + f"Only {', '.join(list(read_func))} files are accepted.")


def detect_encoding(filename):
    """
    Detects the encoding of a file.

    Parameters
    ----------
    filename: str
        Name of the file to detect the encoding.

    Returns
    -------
    encoding: str
        The encoding of the file.

    """
    detector = UniversalDetector()
    with open(filename, 'rb') as file:
        for line in file.readlines():
            detector.feed(line)
    detector.close()
    return detector.result['encoding']


def print_objects_format(object_set, text):
    """
    Return a printable version of the variables in object_sect with the
    header given with text.
    """
    text += " (total %(n_obj)s):\n\t%(objs)s\n" % {
        "n_obj": len(object_set),
        "objs": ", ".join(object_set)
    }
    return text


@dataclass
class Dependencies():
    """
    Representation of variables dependencies.

    Parameters
    ----------
    c_vars: set
        Set of all selected model variables.
    d_deps: dict
        Dictionary of dependencies needed to run vars and modules.
    s_deps: set
        Set of stateful objects to update when integrating selected
        model variables.

    """
    c_vars: Set[str]
    d_deps: Dict[str, set]
    s_deps: Set[str]

    def __str__(self):
        text = print_objects_format(self.c_vars, "Selected variables")

        if self.d_deps["initial"]:
            text += print_objects_format(
                self.d_deps["initial"],
                "\nDependencies for initialization only")
        if self.d_deps["step"]:
            text += print_objects_format(
                self.d_deps["step"],
                "\nDependencies that may change over time")
        if self.d_deps["lookup"]:
            text += print_objects_format(
                self.d_deps["lookup"],
                "\nLookup table dependencies")

        text += print_objects_format(
            self.s_deps,
            "\nStateful objects integrated with the selected variables")

        return text


class ProgressBar:
    """
    Progress bar for integration
    """

    def __init__(self, max_value=None):

        self.max_value = max_value
        if self.max_value is None:
            return

        self.counter = 0

        self.bar = progressbar.ProgressBar(
            max_value=self.max_value,
            widgets=[
                progressbar.ETA(),
                " ",
                progressbar.Bar("#", "[", "]", "-"),
                progressbar.Percentage(),
            ],
        )

        self.bar.start()

    def update(self):
        """Update progress bar"""
        try:
            self.counter += 1
            self.bar.update(self.counter)
        except AttributeError:
            # Error if bar is not imported
            pass

    def finish(self):
        """Finish progress bar"""
        try:
            self.bar.finish()
        except AttributeError:
            # Error if bar is not imported
            pass


class UniqueDims():
    """
    Helper class to create unique dimension names for data_vars with the
    same dimension name but different coords in xarray Datasets.
    """
    def __init__(self, original_dim_name):
        self.dim_name = original_dim_name
        self.dim_prefix = self.dim_name + "_#"
        self.unique_dims = []
        self.num = 1

    def name_new_dim(self, dim_name, coords):
        """
        Returns either a new name (original_dim_name + _# + num) if the coords
        list is not in unique_dims, or the preexisting dimension name if it is.
        Parameters
        ----------
        dim_name: str
            This argument is used to verify that we are passing the right
            dimension name to the class.
        coords: list
            List of coordinates of a dimension.

        Returns
        -------
        Updated name of the original dimension.
        """
        if dim_name != self.dim_name:
            raise ValueError(
                "This object is configured to process dimension "
                f"{self.dim_name} and it's being passed a dimension "
                f"named {dim_name}"
            )
        if self.is_new(coords):
            new_dim_name = self.dim_prefix + str(self.num)
            self.unique_dims.append((new_dim_name, coords))
            self.num += 1
            return new_dim_name
        else:
            for y in self.unique_dims:
                if np.array_equal(coords, y[1]):
                    return y[0]

    def is_new(self, coords):
        """
        Checks if coords is already in the unique_dims list or not.

        Parameters
        ----------
        coords: list
            List of coordinates of a dimension.

        Returns
        -------
        bool
        """
        if not any(
             map(lambda y: np.array_equal(y[1], coords),
                 self.unique_dims)):
            return True
        return False



"""
The Section class allows parsing a model section into Elements. The
final result can be exported to an AbstractSection class in order to
build a model in other language. A section could be either the main model
(without the macros), or a macro definition (not supported yet for Xmile).
"""
from typing import List, Union
from lxml import etree
from pathlib import Path

from ..structures.abstract_model import AbstractSection

from .xmile_element import ControlElement, SubscriptRange, Aux, Flow, Gf, Stock


class Section():
    """
    Section object allows parsing the elements of that section.

    Parameters
    ----------
    name: str
        Section name. '__main__' for the main section or the macro name.

    path: pathlib.Path
        Section path. It should be the model name for main  section and
        the clean macro name for a macro.

    section_type: str ('main' or 'macro')
        The section type.

    params: list
        List of params that takes the section. In the case of main
        section it will be an empty list.

    returns: list
        List of variables that returns the section. In the case of main
        section it will be an empty list.

    content_root: etree._Element
        Section parsed tree content.

    namespace: str
        The namespace of the section given after parsing its content
        with etree.

    split: bool
        If split is True the created section will split the variables
        depending on the views_dict.

    views_dict: dict
        The dictionary of the views. Giving the variables classified at
        any level in order to split them by files.

    """
    _control_vars = ["initial_time", "final_time", "time_step", "saveper"]

    def __init__(self, name: str, path: Path, section_type: str,
                 params: List[str], returns: List[str],
                 content_root: etree._Element, namespace: str, split: bool,
                 views_dict: Union[dict, None]):
        self.name = name
        self.path = path
        self.type = section_type
        self.params = params
        self.returns = returns
        self.content = content_root
        self.ns = {"ns": namespace}
        self.split = split
        self.views_dict = views_dict
        self.elements = None
        self.behaviors = {}

    def __str__(self):  # pragma: no cover
        return "\nSection: %s\n" % self.name

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get section information."""
        text = self.__str__()
        if self.elements:
            for element in self.elements:
                text += element._verbose
        else:
            text += self.content

        return text

    @property
    def verbose(self):  # pragma: no cover
        """Print section information to standard output."""
        print(self._verbose)

    def parse(self, parse_all: bool = True) -> None:
        """
        Parse section object. The subscripts of the section will be added
        to self subscripts. The variables defined as Flows, Auxiliary, Gf,
        and Stock will be converted in XmileElements. The control variables,
        if the section is __main__, will be converted to a ControlElement.

        Parameters
        ----------
        parse_all: bool (optional)
            If True then the created VensimElement objects will be
            automatically parsed. Otherwise, this objects will only be
            added to self.elements but not parser. Default is True.

        """
        # parse subscripts and components
        self.subscripts = self._parse_subscripts()
        self.components = self._parse_components()

        if self.name == "__main__":
            # parse control variables
            self.components += self._parse_control_vars()

        # Parse behavior section
        self.behaviors.update(self._parse_behavior())

        if parse_all:
            [component.parse(self.behaviors) for component in self.components]

        # define elements for printting information
        self.elements = self.subscripts + self.components

    def _parse_behavior(self) -> List[SubscriptRange]:
        """Parse the behavior the section."""
        behaviors = {
            'non_negative_stock': False,
            'non_negative_flow': False
        }
        parsed_bhs = self.content.xpath("ns:behavior", namespaces=self.ns)
        if not parsed_bhs:
            return behaviors

        if parsed_bhs[0].xpath('ns:non_negative', namespaces=self.ns):
            behaviors['non_negative_stock'] = True
            behaviors['non_negative_flow'] = True
            return behaviors

        bhs_stock = parsed_bhs[0].xpath('ns:stock', namespaces=self.ns)
        bhs_flow = parsed_bhs[0].xpath('ns:flow', namespaces=self.ns)

        if bhs_stock and bhs_stock[0].xpath('ns:non_negative', namespaces=self.ns):
            behaviors['non_negative_stock'] = True

        if bhs_flow and bhs_flow[0].xpath('ns:non_negative', namespaces=self.ns):
            behaviors['non_negative_flow'] = True

        return behaviors

    def _parse_subscripts(self) -> List[SubscriptRange]:
        """Parse the subscripts of the section."""
        subscripts = [
            SubscriptRange(
                node.attrib["name"],
                [
                    sub.attrib["name"]
                    for sub in node.xpath("ns:elem", namespaces=self.ns)
                ],
                [])   # no subscript mapping implemented
            for node
            in self.content.xpath("ns:dimensions/ns:dim", namespaces=self.ns)
        ]
        self.subscripts_dict = {
            subr.name: subr.definition for subr in subscripts}
        return subscripts

    def _parse_components(self) -> List[Union[Flow, Aux, Gf, Stock]]:
        """
        Parse model components. Four groups defined:
        Aux: auxiliary variables
        Flow: flows
        Gf: lookups
        Stock: integs

        """
        # Add auxiliary variables
        components = [
            Aux(node, self.ns, self.subscripts_dict)
            for node in self.content.xpath(
                "ns:model/ns:variables/ns:aux",
                namespaces=self.ns)
            if node.attrib["name"].lower().replace(" ", "_")
            not in self._control_vars]

        # Add flows
        components += [
            Flow(node, self.ns, self.subscripts_dict)
            for node in self.content.xpath(
                "ns:model/ns:variables/ns:flow",
                namespaces=self.ns)
            if node.attrib["name"].lower().replace(" ", "_")
            not in self._control_vars]

        # Add lookups
        components += [
            Gf(node, self.ns, self.subscripts_dict)
            for node in self.content.xpath(
                "ns:model/ns:variables/ns:gf",
                namespaces=self.ns)
            ]

        # Add stocks
        components += [
            Stock(node, self.ns, self.subscripts_dict)
            for node in self.content.xpath(
                "ns:model/ns:variables/ns:stock",
                namespaces=self.ns)
            ]

        return components

    def _parse_control_vars(self) -> List[ControlElement]:
        """Parse control vars and rename them with Vensim standard."""

        # Read the start time of simulation
        node = self.content.xpath('ns:sim_specs', namespaces=self.ns)[0]
        time_units = node.attrib['time_units']\
            if 'time_units' in node.attrib else ""

        control_vars = []

        # initial time of the simulation
        control_vars.append(ControlElement(
            name="INITIAL TIME",
            units=time_units,
            documentation="The initial time for the simulation.",
            eqn=node.xpath("ns:start", namespaces=self.ns)[0].text
        ))

        # final time of the simulation
        control_vars.append(ControlElement(
            name="FINAL TIME",
            units=time_units,
            documentation="The final time for the simulation.",
            eqn=node.xpath("ns:stop", namespaces=self.ns)[0].text
        ))

        # time step of simulation
        dt_node = node.xpath("ns:dt", namespaces=self.ns)[0]
        dt_eqn = "1/(" + dt_node.text + ")" if "reciprocal" in dt_node.attrib\
            and dt_node.attrib["reciprocal"].lower() == "true"\
            else dt_node.text

        control_vars.append(ControlElement(
            name="TIME STEP",
            units=time_units,
            documentation="The time step for the simulation.",
            eqn=dt_eqn
        ))

        # saving time of the simulation = time step
        control_vars.append(ControlElement(
            name="SAVEPER",
            units=time_units,
            documentation="The save time step for the simulation.",
            eqn="time_step"
        ))

        return control_vars

    def get_abstract_section(self) -> AbstractSection:
        """
        Get Abstract Section used for building. This, method should be
        called after parsing the section (self.parse). This method is
        automatically called by Model's get_abstract_model and
        automatically generates the AbstractSubscript ranges and merge
        the components in elements calling also the get_abstract_components
        method from each model component.

        Returns
        -------
        AbstractSection: AbstractSection
          Abstract Section object that can be used for building the model
          in another language.

        """
        return AbstractSection(
            name=self.name,
            path=self.path,
            type=self.type,
            params=self.params,
            returns=self.returns,
            subscripts=[
                subs_range.get_abstract_subscript_range()
                for subs_range in self.subscripts
            ],
            elements=[
                element.get_abstract_element()
                for element in self.components
            ],
            constraints=[],
            test_inputs=[],
            split=self.split,
            views_dict=self.views_dict
        )



"""
Macro and Model classes are the main classes for loading and interacting
with a PySD model. Model class allows loading and running a PySD model.
Several methods and propierties are inherited from Macro class, which
allows integrating a model or a Macro expression (set of functions in
a separate file).
"""
import time
import warnings
import inspect
import pickle
from pathlib import Path
from copy import deepcopy
from typing import Union

import numpy as np
import xarray as xr
import pandas as pd

from pysd._version import __version__

from . import utils
from .statefuls import DynamicStateful, Stateful
from .external import External, Excels, ExtLookup, ExtData

from .cache import Cache, constant_cache
from .data import TabData
from .lookups import HardcodedLookups
from .components import Components, Time
from .output import ModelOutput


class Macro(DynamicStateful):
    """
    The Macro class implements a stateful representation of the system,
    and contains the majority of methods for accessing and modifying
    components.

    When the instance in question also serves as the root model object
    (as opposed to a macro or submodel within another model) it will have
    added methods to facilitate execution.

    The Macro object will be created with components drawn from a
    translated Python model file.

    Parameters
    ----------
    py_model_file: str or pathlib.Path
        Filename of a model or macro which has already been converted
        into a Python format.
    params: dict or None (optional)
        Dictionary of the macro parameters. Default is None.
    return_func: str or None (optional)
        The name of the function to return from the macro. Default is None.
    time: components.Time or None (optional)
        Time object for integration. If None a new time object will
        be generated (for models), if passed the time object will be
        used (for macros). Default is None.
    time_initialization: callable or None
        Time to set at the begginning of the Macro. Default is None.
    data_files: dict or list or str or None
        The dictionary with keys the name of file and variables to
        load the data from. Or the list of names or name of the file
        to search the data in. Only works for TabData type object
        and it is neccessary to provide it. Default is None.
    py_name: str or None
        The name of the Macro object. Default is None.

    See also
    --------
    :class:`pysd.py_backend.model.Model`

    """
    def __init__(self, py_model_file, params=None, return_func=None,
                 time=None, time_initialization=None, data_files=None,
                 py_name=None):
        super().__init__()
        self.time = time
        self.time_initialization = time_initialization
        # Initialize the cache object
        self.cache = Cache()
        # Python name of the object (for Macros)
        self.py_name = py_name
        # Booleans to avoid loading again external data or lookups
        self.external_loaded = False
        self.lookups_loaded = False
        # Functions with constant cache
        self._constant_funcs = set()
        # Attributes that are set later
        self.stateful_initial_dependencies = None
        self.initialize_order = None
        self.cache_type = None
        self._components_setter_tracker = {}
        # Load model/macro from file and save in components
        self.components = Components(str(py_model_file), self.set_components)

        if __version__.split(".")[0]\
           != self.get_pysd_compiler_version().split(".")[0]:
            raise ImportError(
                "\n\nNot able to import the model. "
                + "The model was translated with a "
                + "not compatible version of PySD:"
                + "\n\tPySD " + self.get_pysd_compiler_version()
                + "\n\nThe current version of PySd is:"
                + "\n\tPySD " + __version__ + "\n\n"
                + "Please translate again the model with the function"
                + " read_vensim or read_xmile.")

        # Assing some protected attributes for easier access
        self._namespace = self.components._components.component.namespace
        self._dependencies =\
            self.components._components.component.dependencies.copy()
        self._subscript_dict = getattr(
            self.components._components, "_subscript_dict", {})
        self._modules = getattr(
            self.components._components, "_modules", {})

        self._doc = self._build_doc()

        if params is not None:
            # add params to namespace
            self._namespace.update(self.components._components._params)
            # create new components with the params
            self._set_components(params, new=True)
            # update dependencies
            for param in params:
                self._dependencies[
                    self._namespace[param]] = {"time"}

        # Get the collections of stateful elements and external elements
        self._stateful_elements = {
            name: getattr(self.components, name)
            for name in dir(self.components)
            if isinstance(getattr(self.components, name), Stateful)
        }
        self._dynamicstateful_elements = [
            getattr(self.components, name) for name in dir(self.components)
            if isinstance(getattr(self.components, name), DynamicStateful)
        ]
        self._external_elements = [
            getattr(self.components, name) for name in dir(self.components)
            if isinstance(getattr(self.components, name), External)
        ]
        self._macro_elements = [
            getattr(self.components, name) for name in dir(self.components)
            if isinstance(getattr(self.components, name), Macro)
        ]

        self._data_elements = [
            getattr(self.components, name) for name in dir(self.components)
            if isinstance(getattr(self.components, name), TabData)
        ]

        self._lookup_elements = [
            getattr(self.components, name) for name in dir(self.components)
            if isinstance(getattr(self.components, name), HardcodedLookups)
        ]

        # Load data files
        if data_files:
            self._get_data(data_files)

        # Assign the cache type to each variable
        self._assign_cache_type()
        # Get the initialization order of Stateful elements
        self._get_initialize_order()

        if return_func is not None:
            # Assign the return value of Macros
            self.return_func = getattr(self.components, return_func)
        else:
            self.return_func = lambda: 0

        self.py_model_file = str(py_model_file)

    def __call__(self):
        return self.return_func()

    @property
    def doc(self) -> pd.DataFrame:
        """
        The documentation of the model.
        """
        return self._doc.copy()

    @property
    def namespace(self) -> dict:
        """
        The namespace dictionary of the model.
        """
        return self._namespace.copy()

    @property
    def dependencies(self) -> dict:
        """
        The dependencies dictionary of the model.
        """
        return self._dependencies.copy()

    @property
    def subscripts(self) -> dict:
        """
        The subscripts dictionary of the model.
        """
        return self._subscript_dict.copy()

    @property
    def modules(self) -> Union[dict, None]:
        """
        The dictionary of modules of the model. If the model is not
        split by modules it returns None.
        """
        return self._modules.copy() or None

    def clean_caches(self):
        """
        Clean the cache of the object and the macros objects that it
        contains
        """
        self.cache.clean()
        # if nested macros
        [macro.clean_caches() for macro in self._macro_elements]

    def _get_data(self, data_files):
        if isinstance(data_files, dict):
            for data_file, vars in data_files.items():
                for var in vars:
                    found = False
                    for element in self._data_elements:
                        if var in [element.py_name, element.real_name]:
                            element.load_data(data_file)
                            found = True
                            break
                    if not found:
                        raise ValueError(
                            f"'{var}' not found as model data variable")

        else:
            for element in self._data_elements:
                element.load_data(data_files)

    def _get_initialize_order(self):
        """
        Get the initialization order of the stateful elements
        and their the full dependencies.
        """
        # get the full set of dependencies to initialize an stateful object
        # includying all levels
        self.stateful_initial_dependencies = {
            ext: set()
            for ext in self._dependencies
            if (ext.startswith("_") and not ext.startswith("_active_initial_"))
        }
        for element in self.stateful_initial_dependencies:
            self._get_full_dependencies(
                element, self.stateful_initial_dependencies[element],
                "initial")

        # get the full dependencies of stateful objects taking into account
        # only other objects
        current_deps = {
            element: [
                dep for dep in deps
                if dep in self.stateful_initial_dependencies
            ] for element, deps in self.stateful_initial_dependencies.items()
        }

        # get initialization order of the stateful elements
        self.initialize_order = []
        delete = True
        while delete:
            delete = []
            for element in current_deps:
                if not current_deps[element]:
                    # if stateful element has no deps on others
                    # add to the queue to initialize
                    self.initialize_order.append(element)
                    delete.append(element)
                    for element2 in current_deps:
                        # remove dependency on the initialized element
                        if element in current_deps[element2]:
                            current_deps[element2].remove(element)
            # delete visited elements
            for element in delete:
                del current_deps[element]

        if current_deps:
            # if current_deps is not an empty set there is a circular
            # reference between stateful objects
            raise ValueError(
                'Circular initialization...\n'
                + 'Not able to initialize the following objects:\n\t'
                + '\n\t'.join(current_deps))

    def _get_full_dependencies(self, element, dep_set, stateful_deps):
        """
        Get all dependencies of an element, i.e., also get the dependencies
        of the dependencies. When finding an stateful element only dependencies
        for initialization are considered.

        Parameters
        ----------
        element: str
            Element to get the full dependencies.
        dep_set: set
            Set to include the dependencies of the element.
        stateful_deps: "initial" or "step"
            The type of dependencies to take in the case of stateful objects.

        Returns
        -------
        None

        """
        deps = self._dependencies[element]
        if element.startswith("_"):
            deps = deps[stateful_deps]
        for dep in deps:
            if dep not in dep_set and not dep.startswith("__")\
               and dep != "time":
                dep_set.add(dep)
                self._get_full_dependencies(dep, dep_set, stateful_deps)

    def _add_constant_cache(self):
        for element, cache_type in self.cache_type.items():
            if cache_type == "run":
                self.components._set_component(
                    element,
                    constant_cache(getattr(self.components, element))
                )
                self._constant_funcs.add(element)

    def _remove_constant_cache(self):
        for element in self._constant_funcs:
            self.components._set_component(
                element,
                getattr(self.components, element).function)
            # remove attributes added with constant cache
            delattr(getattr(self.components, element), 'function')
            delattr(getattr(self.components, element), 'value')
        self._constant_funcs.clear()

    def _assign_cache_type(self):
        """
        Assigns the cache type to all the elements from the namespace.
        """
        self.cache_type = {"time": None}

        for element in self._namespace.values():
            if element not in self.cache_type\
               and element in self._dependencies:
                self._assign_cache(element)

        for element, cache_type in self.cache_type.items():
            if cache_type is not None:
                if element not in self.cache.cached_funcs\
                   and self._count_calls(element) > 1:
                    self.components._set_component(
                        element,
                        self.cache(getattr(self.components, element)))
                    self.cache.cached_funcs.add(element)

    def _count_calls(self, element):
        n_calls = 0
        for subelement in self._dependencies:
            if subelement.startswith("_") and\
               element in self._dependencies[subelement]["step"]:
                if element in\
                   self._dependencies[subelement]["initial"]:
                    n_calls +=\
                        2*self._dependencies[subelement]["step"][element]
                else:
                    n_calls +=\
                        self._dependencies[subelement]["step"][element]
            elif (not subelement.startswith("_") and
                  element in self._dependencies[subelement]):
                n_calls +=\
                    self._dependencies[subelement][element]

        return n_calls

    def _assign_cache(self, element):
        """
        Assigns the cache type to the given element and its dependencies if
        needed.

        Parameters
        ----------
        element: str
            Element name.

        Returns
        -------
        None

        """
        if not self._dependencies[element]:
            self.cache_type[element] = "run"
        elif "__lookup__" in self._dependencies[element]:
            self.cache_type[element] = None
        elif self._isdynamic(self._dependencies[element]):
            self.cache_type[element] = "step"
        else:
            self.cache_type[element] = "run"
            for subelement in self._dependencies[element]:
                if subelement.startswith("_initial_")\
                   or subelement.startswith("__"):
                    continue
                if subelement not in self.cache_type:
                    self._assign_cache(subelement)
                if self.cache_type[subelement] == "step":
                    self.cache_type[element] = "step"
                    break

    def _isdynamic(self, dependencies):
        """

        Parameters
        ----------
        dependencies: iterable
            List of dependencies.

        Returns
        -------
        isdynamic: bool
            True if 'time' or a dynamic stateful objects is in dependencies.

        """
        if "time" in dependencies:
            return True
        for dep in dependencies:
            if dep.startswith("_") and not dep.startswith("_initial_")\
               and not dep.startswith("__"):
                return True
        return False

    def get_pysd_compiler_version(self):
        """
        Returns the version of pysd complier that used for generating
        this model
        """
        return self.components.__pysd_version__

    def initialize(self):
        """
        This function initializes the external objects and stateful objects
        in the given order.
        """
        # Initialize time
        if self.time is None:
            self.time = self.time_initialization()

        # Reset time to the initial one
        self.time.reset()
        self.cache.clean()

        self.components._init_outer_references({
            'scope': self,
            'time': self.time
        })

        if not self.lookups_loaded:
            # Initialize HardcodedLookups elements
            for element in self._lookup_elements:
                element.initialize()

            self.lookups_loaded = True

        if not self.external_loaded:
            # Initialize external elements
            self.initialize_external_data()

        # Initialize stateful objects
        for element_name in self.initialize_order:
            self._stateful_elements[element_name].initialize()

    def ddt(self):
        return np.array([component.ddt() for component
                         in self._dynamicstateful_elements], dtype=object)

    @property
    def state(self):
        return np.array([component.state for component
                         in self._dynamicstateful_elements], dtype=object)

    @state.setter
    def state(self, new_value):
        [component.update(val) for component, val
         in zip(self._dynamicstateful_elements, new_value)]

    def initialize_external_data(self, externals=None):

        """
        Initializes external data.

        If a path to a netCDF file containing serialized values of some or
        all of the model external data is passed in the external argument,
        those will be loaded from the file.

        To get the full performance gain of loading the externals from a netCDF
        file, the model should be loaded with initialize=False first.

        Examples of usage are available at
        `Advanced Usage <https://pysd.readthedocs.io/en/master/advanced_usage.html#initializing-external-data-from-netcdf-file>`__.

        Parameters
        ----------
        externals: str or pathlib.Path (optional)
            Path to the netCDF file that contains the model external objects.

        Returns
        -------
        None

        See also
        --------
        :func:`pysd.py_backend.model.Macro.serialize_externals`

        Note
        ----
        To load externals from a netCDF file you need to have installed
        the optional dependency `netCDF4`.

        """

        if not externals:
            for ext in self._external_elements:
                ext.initialize()

            # Remove Excel data from memory
            Excels.clean()

            self.external_loaded = True

            return

        externals = Path(externals)

        if not externals.is_file():
            raise FileNotFoundError(f"Invalid file path ({str(externals)})")

        try:
            ds = xr.open_dataset(externals)
        except ValueError:  # pragma: no cover
            raise ModuleNotFoundError("No module named 'netCDF4'")

        for ext in self._external_elements:
            if ext.py_name in ds.data_vars.keys():
                # Initialize external from nc file
                da = ds.data_vars[ext.py_name]
                if isinstance(ext, ExtData):
                    # Rename again time dimension
                    time_dim = [dim for dim in da.dims if dim.startswith(
                        "time_#")][0]
                    da = da.rename({time_dim: "time"})
                elif isinstance(ext, ExtLookup):
                    # Rename again lookup_dim dimension
                    lookup_dim = [dim for dim in da.dims if dim.startswith(
                        "lookup_dim_#")][0]
                    da = da.rename({lookup_dim: "lookup_dim"})
                # Assign the value
                if da.dims:
                    ext.data = da
                else:
                    ext.data = float(da.data)
            else:
                # Initialize external from original file
                ext.initialize()

        Excels.clean()

        self.external_loaded = True

    def serialize_externals(self, export_path="externals.nc",
                            include_externals="all", exclude_externals=None):
        """
        Stores a netCDF file with the data and metadata for all model external
        objects.

        This method is useful for models with lots of external inputs, which
        are slow to load into memory. Once exported, the resulting netCDF file
        can be passed as argument to the initialize_external_data method.

        Names of variables should be those in the model (python safe,
        without the _ext_type_ string in front).

        Examples of usage are available at
        `Advanced Usage <https://pysd.readthedocs.io/en/master/advanced_usage.html#initializing-external-data-from-netcdf-file>`__.

        Parameters
        ----------
        export_path: str or pathlib.Path (optional)
            Path of the resulting *.nc* file.
        include_externals: list or str (optional)
            External objects to export to netCDF.
            If 'all', then all externals are exported to the *.nc* file.
            The argument also accepts a list containing spreadsheet file names,
            external variable names or a combination of both. If a spreadsheet
            file path is passed, all external objects defined in it will be
            included in the *.nc* file. Spreadsheet tab names are not currently
            supported, because the same may be used in different files.
            Better customisation can be achieved by combining the
            include_externals and exclude_externals (see description below)
            arguments.
        exclude_externals: list or None (optional)
            Exclude external objects from being included in the exported nc
            file. It accepts either variable names, spreadsheet files or a
            combination of both.

        Returns
        -------
        None

        See also
        --------
        :func:`pysd.py_backend.model.Macro.initialize_external_data`

        Note
        ----
        To run this function you need to have installed the optional
        dependency `netCDF4`.

        """
        data = {}
        metadata = {}

        lookup_dims = utils.UniqueDims("lookup_dim")
        data_dims = utils.UniqueDims("time")

        if isinstance(export_path, str):
            export_path = Path(export_path)

        if not include_externals:
            raise ValueError("include_externals argument must not be None.")

        # Generate a Dataframe to make simpler the search of external
        # objects to include
        # TODO include also checking the original name
        py_names = []
        externals_dict = {"py_var_name": [], "file": [], "ext": []}
        for ext in self._external_elements:
            py_names.append(ext.py_name)
            externals_dict["py_var_name"].append(
                self.__get_varname_from_ext_name(ext.py_name))
            externals_dict["file"].append(set(ext.files))
            externals_dict["ext"].append(ext)
        exts_df = pd.DataFrame(index=py_names, data=externals_dict)

        if include_externals != "all":
            if not isinstance(include_externals, (list, set)):
                raise TypeError(
                    "include_externals must be 'all', or a list, or a set.")

            # subset only to the externals to include
            exts_df = exts_df[[
                name in include_externals
                or var_name in include_externals
                or bool(file.intersection(include_externals))
                for name, (var_name, file)
                in exts_df[["py_var_name", "file"]].iterrows()
            ]]

        if exclude_externals:
            if not isinstance(exclude_externals, (list, set)):
                raise TypeError("exclude_externals must be a list or a set.")

            # subset only to the externals to include
            exts_df = exts_df[[
                name not in exclude_externals
                and var_name not in exclude_externals
                and not bool(file.intersection(exclude_externals))
                for name, (var_name, file)
                in exts_df[["py_var_name", "file"]].iterrows()
            ]]

        for _, (ext, var_name) in exts_df[["ext", "py_var_name"]].iterrows():
            self.__include_for_serialization(
                ext, var_name, data, metadata, lookup_dims, data_dims
            )

        # create description to be used as global attribute of the dataset
        description = {
            "description": f"External objects for {self.py_model_file} "
                           f"exported on {time.ctime(time.time())} "
                           f"using PySD version {__version__}"
        }

        # create Dataset
        ds = xr.Dataset(data_vars=data, attrs=description)

        # add data_vars attributes
        for key, values in metadata.items():
            ds[key].attrs = values

        try:
            ds.to_netcdf(export_path)
        except KeyError:  # pragma: no cover
            raise ModuleNotFoundError("No module named 'netCDF4'")

    def __include_for_serialization(self, ext, py_name_clean, data, metadata,
                                    lookup_dims, data_dims):
        """
        Initialize the external object and get the data and metadata for
        inclusion in the netCDF.

        This function updates the metadata dict with the metadata corresponding
        to each external object, which is collected from model.doc. It also
        updates the data dict, with the data from the external object (ext).

        It renames the "time" dimension of ExtData types by appending _#
        followed by a unique number at the end of it. For large models, this
        prevents having an unnecessary large time dimension, which then causes
        all ExtData objects to have many nans when stored in a xarray Dataset.
        It does the same for the "lookup_dim" of all ExtLookup objects.

        Note
        ----
        Though subscripts can be read from Excel, they are hardcoded
        during the model building process. Therefore they will not be
        serialized.

        Parameters
        ----------
        ext: pysd.py_backend.externals.External
            External object. It can be any of the External subclasses
            (ExtConstant, ExtData, ExtLookup)
        py_name_clean: str
            Name of the variable without _ext_[constant|data|lookup] prefix.
        data: dict
            Collects all the data for each external, which is later used to
            build the xarray Dataset.
        metadata: dict
            Collects the metadata for each external, which is later included
            as data_vars attributes in the xarray Dataset.
        lookup_dims: utils.UniqueDims
            UniqueDims object for "lookup_dim" dimension.
        data_dims: utils.UniqueDims
            UniqueDims object for "time" dimension.

        Returns
        -------
        None

        """
        ext.initialize()

        # collecting variable metadata from model._doc
        var_meta = {
            col:
            self._doc.loc[self._doc["Py Name"] == py_name_clean, col].values[0]
            or "Missing"
            for col in self._doc.columns
        }
        var_meta["files"] = ";".join(ext.files)
        var_meta["tabs"] = ";".join(ext.tabs)
        var_meta["cells"] = ";".join(ext.cells)
        # TODO: add also time_row_or_cols

        da = ext.data

        # Renaming shared dims by all ExtData ("time") and ExtLookup
        # ("lookup_dim") external objects
        if isinstance(ext, ExtData):
            new_name = data_dims.name_new_dim("time", da.coords["time"].values)
            da = da.rename({"time": new_name})
        if isinstance(ext, ExtLookup):
            new_name = lookup_dims.name_new_dim("lookup_dim",
                                                da.coords["lookup_dim"].values)
            da = da.rename({"lookup_dim": new_name})

        metadata.update({ext.py_name: var_meta})
        data.update({ext.py_name: da})

        # TODO use a logger
        print(f"Finished processing variable {py_name_clean}.")

    def __get_varname_from_ext_name(self, varname):
        """
        Returns the name of the variable that depends on the external object
        named varname. If that is not possible (see warning in the code to
        understand when that may happen), it gets the name by removing the
        _ext_[constant|data|lookup] prefix from the varname.

        Parameters
        ----------
        varname: str
            Variable name to which the External object is assigned.

        Returns
        -------
        var: str
            Name of the variable that calls the variable with name varname.

        """
        for var, deps in self._dependencies.items():
            for _, ext_name in deps.items():
                if varname == ext_name:
                    return var

        warnings.warn(
            f"No variable depends upon '{varname}'. This is likely due "
            f"to the fact that '{varname}' is defined using a mix of "
            "DATA and CONSTANT. Though Vensim allows it, it is "
            "not recommended."
        )
        return "_".join(varname.split("_")[3:])

    def get_args(self, param):
        """
        Returns the arguments of a model element.

        Parameters
        ----------
        param: str or func
            The model element name or function.

        Returns
        -------
        args: list
            List of arguments of the function.

        Examples
        --------
        >>> model.get_args('birth_rate')
        >>> model.get_args('Birth Rate')

        See also
        --------
        :func:`pysd.py_backend.model.Macro.get_coords`

        """
        if isinstance(param, str):
            func_name = utils.get_key_and_value_by_insensitive_key_or_value(
                param,
                self._namespace)[1] or param

            func = getattr(self.components, func_name)
        else:
            func = param

        if hasattr(func, 'args'):
            # cached functions
            return func.args
        else:
            # regular functions
            args = inspect.getfullargspec(func)[0]
            if 'self' in args:
                args.remove('self')
            return args

    def get_coords(self, param):
        """
        Returns the coordinates and dims of a model element.

        Parameters
        ----------
        param: str or func
            The model element name or function.

        Returns
        -------
        (coords, dims) or None: (dict, list) or None
            The coords and the dimensions of the element if it has.
            Otherwise, returns None.

        Examples
        --------
        >>> model.get_coords('birth_rate')
        >>> model.get_coords('Birth Rate')

        See also
        --------
        :func:`pysd.py_backend.model.Macro.get_args`

        """
        if isinstance(param, str):
            func_name = utils.get_key_and_value_by_insensitive_key_or_value(
                param,
                self._namespace)[1] or param

            func = getattr(self.components, func_name)

        else:
            func = param

        if hasattr(func, "subscripts"):
            dims = func.subscripts
            if not dims:
                return None
            coords = {dim: self.components._subscript_dict[dim]
                      for dim in dims}
            return coords, dims
        elif hasattr(func, "state") and isinstance(func.state, xr.DataArray):
            value = func()
        else:
            return None

        dims = list(value.dims)
        coords = {coord: list(value.coords[coord].values)
                  for coord in value.coords}
        return coords, dims

    def __getitem__(self, param):
        """
        Returns the current value of a model component.

        Parameters
        ----------
        param: str or func
            The model element name.

        Returns
        -------
        value: float or xarray.DataArray
            The value of the model component.

        Examples
        --------
        >>> model['birth_rate']
        >>> model['Birth Rate']

        Note
        ----
        It will crash if the model component takes arguments.

        See also
        --------
        :func:`pysd.py_backend.model.Macro.get_series_data`

        """
        func_name = utils.get_key_and_value_by_insensitive_key_or_value(
            param,
            self._namespace)[1] or param

        if self.get_args(getattr(self.components, func_name)):
            raise ValueError(
                "Trying to get the current value of a lookup "
                "to get all the values with the series data use "
                "model.get_series_data(param)\n\n")

        return getattr(self.components, func_name)()

    def get_series_data(self, param):
        """
        Returns the original values of a model lookup/data component.

        Parameters
        ----------
        param: str
            The model lookup/data element name.

        Returns
        -------
        value: xarray.DataArray
            Array with the value of the interpolating series
            in the first dimension.

        Examples
        --------
        >>> model['room_temperature']
        >>> model['Room temperature']

        """
        func_name = utils.get_key_and_value_by_insensitive_key_or_value(
            param,
            self._namespace)[1] or param

        if func_name.startswith("_ext_"):
            return getattr(self.components, func_name).data
        elif "__data__" in self._dependencies[func_name]:
            return getattr(
                self.components,
                self._dependencies[func_name]["__data__"]
            ).data
        elif "__lookup__" in self._dependencies[func_name]:
            return getattr(
                self.components,
                self._dependencies[func_name]["__lookup__"]
            ).data
        else:
            raise ValueError(
                "Trying to get the values of a constant variable. "
                "'model.get_series_data' only works lookups/data objects.\n\n")

    def set_components(self, params):
        """
        Set the value of exogenous model elements.
        Element values should be passed with a dictionary in the
        function call. Values can be numeric type or pandas Series.
        Series will be interpolated by integrator.

        Parameters
        ----------
        params: dict
            Dictionary with the name of the elements to modify and the
            value that they would take. If the passed value is a
            :class:`float` or a :class:`xarray.DataArray` (must be
            compatible with the dimensions of the variable). In this
            case, the variable will have a constant value, returning
            the past value and broadcasting to all dimensions, if
            necessary. If a :class:`pandas.Series` is passed, the
            variable will be of type data and will use the time of the
            model to interpolate the result having as reference the
            indexes of the series. In this case, the series 'data' can
            also be a :class:`float` or a :class:`xarray.DataArray`,
            as with constant values. In the case of the target using
            a :class:`pysd.py_backend.lookups.Lookup` object, it will
            modify the object values to use the original arguments when
            being call. More detailed information and examples of usage
            are available at
            `Getting Started <https://pysd.readthedocs.io/en/master/getting_started.html#setting-parameter-values>`__.

            To write more complex relationships, which may or may not
            include other model variables, a callable, e.g. a function,
            can be passed that takes the same arguments as the original
            function and returns a :class:`float` or a
            :class:`xarray.DataArray` with exactly the same dimensions
            as the original function. More detailed information and
            examples of usage are available at
            `Advanced Usage <https://pysd.readthedocs.io/en/master/advanced_usage.html#replacing-model-components-with-more-complex-objects>`__.


        Note
        ----
        This function is to modify the value or equations of the
        variables, it won't work properly with Stateful objects, e.g.
        Integ, DelayFixed... In order to modify them it will be
        necessary to do it manually, if you have other inputs it is
        recommended to modify these ones. To change their initial
        value :func:`pysd.py_backend.model.Model.set_initial_condition`
        method could be used.


        Examples
        --------
        >>> model.set_components({'birth_rate': 10})
        >>> model.set_components({'Birth Rate': 10})

        >>> br = pandas.Series(index=range(30), data=np.sin(range(30))
        >>> model.set_components({'birth_rate': br})

        See also
        --------
        :func:`pysd.py_backend.model.Model.set_initial_condition`
        :func:`pysd.py_backend.model.Macro.get_coords`
        :func:`pysd.py_backend.model.Macro.get_args`

        """
        self._components_setter_tracker.update(params)
        self._set_components(params, new=False)

    def _set_components(self, params, new):
        """
        Set the value of exogenous model elements, giving the option to
        set new components (used in Macros).
        """

        for key, value in params.items():
            func_name = utils.get_key_and_value_by_insensitive_key_or_value(
                key,
                self._namespace)[1]

            if isinstance(value, np.ndarray) or isinstance(value, list):
                raise TypeError(
                    'When setting ' + key + '\n'
                    'Setting subscripted must be done using a xarray.DataArray'
                    ' with the correct dimensions or a constant value '
                    '(https://pysd.readthedocs.io/en/master/'
                    'getting_started.html)')

            if func_name is None:
                raise NameError(
                    "\n'%s' is not recognized as a model component."
                    % key)

            if new:
                func = None
                dims = None
            else:
                func = getattr(self.components, func_name)
                _, dims = self.get_coords(func) or (None, None)

            # if the variable is a lookup or a data we perform the change in
            # the object they call
            func_type = getattr(func, "type", None)
            if func_type in ["Lookup", "Data"]:
                # getting the object from original dependencies
                obj = self._dependencies[func_name][f"__{func_type.lower()}__"]
                getattr(
                    self.components,
                    obj
                ).set_values(value)

                if not isinstance(value, pd.Series):
                    warnings.warn(
                        "Replacing interpolation data with constant values.")

                # Update dependencies
                if func_type == "Data":
                    if isinstance(value, pd.Series):
                        self._dependencies[func_name] = {
                            "time": 1, "__data__": obj
                        }
                    else:
                        self._dependencies[func_name] = {"__data__": obj}

                continue

            if func_type == "Stateful":
                warnings.warn(
                    "Replacing the value of Stateful variable with "
                    "an expression. To set initial conditions use "
                    "`set_initial_condition` instead..."
                )

            if isinstance(value, pd.Series):
                if func_type == "Constant":
                    warnings.warn(
                        "Replacing a constant value with a "
                        "time-dependent value. The value will be "
                        "interpolated over time."
                    )
                new_function, deps = self._timeseries_component(
                    value, dims)
                self._dependencies[func_name] = deps
            elif callable(value):
                if func_type == "Constant":
                    warnings.warn(
                        "Replacing a constant value with a callable. "
                        "The value may not be constant anymore."
                    )
                new_function = value
                # Using step cache adding time as dependency
                # TODO it would be better if we can parse the content
                # of the function to get all the dependencies
                self._dependencies[func_name] = {"time": 1}

            else:
                if func_type != "Constant":
                    warnings.warn("Replacing a variable by a constant value.")
                new_function = self._constant_component(value, dims)
                self._dependencies[func_name] = {}

            # copy attributes from the original object to proper working
            # of internal functions
            new_function.__name__ = func_name
            new_function.__dict__.update(getattr(func, "__dict__", {}))
            # set the new function
            self.components._set_component(func_name, new_function)
            if func_name in self.cache.cached_funcs:
                self.cache.cached_funcs.remove(func_name)

    def _timeseries_component(self, series, dims):
        """ Internal function for creating a timeseries model element """
        # this is only called if the set_component function recognizes a
        # pandas series
        # TODO: raise a warning if extrapolating from the end of the series.
        # TODO: data type variables should be creted using a Data object
        # lookup type variables should be created using a Lookup object

        if isinstance(series.values[0], xr.DataArray):
            # the interpolation will be time dependent
            return lambda: utils.rearrange(xr.concat(
                series.values,
                series.index).interp(concat_dim=self.time()).reset_coords(
                'concat_dim', drop=True),
                dims, self._subscript_dict), {'time': 1}

        elif dims:
            # the interpolation will be time dependent
            return lambda: utils.rearrange(
                np.interp(self.time(), series.index, series.values),
                dims, self._subscript_dict), {'time': 1}

        else:
            # the interpolation will be time dependent
            return lambda: np.interp(
                self.time(), series.index, series.values
                ), {'time': 1}

    def _constant_component(self, value, dims):
        """ Internal function for creating a constant model element """
        if dims:
            return lambda: utils.rearrange(
                value, dims, self._subscript_dict)

        else:
            return lambda: value

    def set_initial_value(self, time, initial_value):
        """
        Set the system initial value.

        Parameters
        ----------
        time : float or int
            The system intial time to be set.

        initial_value : dict
            A (possibly partial) dictionary of the system initial values.
            The keys to this dictionary may be either pysafe names or
            original model file names.

        See also
        --------
        :func:`pysd.py_backend.model.Model.set_initial_condition`

        """
        self.time.set_control_vars(initial_time=time)
        stateful_name = "_NONE"
        modified_statefuls = set()

        for key, value in initial_value.items():
            component_name =\
                utils.get_key_and_value_by_insensitive_key_or_value(
                    key, self._namespace)[1]
            if component_name is not None:
                if self._dependencies[component_name]:
                    deps = list(self._dependencies[component_name])
                    if len(deps) == 1 and deps[0] in self.initialize_order:
                        stateful_name = deps[0]
            else:
                component_name = key
                stateful_name = key

            try:
                _, dims = self.get_coords(component_name)
            except TypeError:
                dims = None

            if isinstance(value, xr.DataArray)\
               and not set(value.dims).issubset(set(dims)):
                raise ValueError(
                    f"\nInvalid dimensions for {component_name}."
                    f"It should be a subset of {dims}, "
                    f"but passed value has {list(value.dims)}")

            if isinstance(value, np.ndarray) or isinstance(value, list):
                raise TypeError(
                    'When setting ' + key + '\n'
                    'Setting subscripted must be done using a xarray.DataArray'
                    ' with the correct dimensions or a constant value '
                    '(https://pysd.readthedocs.io/en/master/'
                    'getting_started.html)')

            # Try to update stateful component
            try:
                element = getattr(self.components, stateful_name)
                if dims:
                    value = utils.rearrange(
                        value, dims,
                        self._subscript_dict)
                element.initialize(value)
                modified_statefuls.add(stateful_name)
            except NameError:
                # Try to override component
                raise ValueError(
                    f"\nUnrecognized stateful '{component_name}'. If you want"
                    " to set a value of a regular component. Use params={"
                    f"'{component_name}': {value}" + "} instead.")

        self.clean_caches()

        # get the elements to initialize
        elements_to_initialize =\
            self._get_elements_to_initialize(modified_statefuls)

        # Initialize remaining stateful objects
        for element_name in self.initialize_order:
            if element_name in elements_to_initialize:
                self._stateful_elements[element_name].initialize()

    def _get_elements_to_initialize(self, modified_statefuls):
        elements_to_initialize = set()
        for stateful, deps in self.stateful_initial_dependencies.items():
            if stateful in modified_statefuls:
                # if elements initial conditions have been modified
                # we should not modify it
                continue
            for modified_sateteful in modified_statefuls:
                if modified_sateteful in deps:
                    # if element has dependencies on a modified element
                    # we should re-initialize it
                    elements_to_initialize.add(stateful)
                    continue

        return elements_to_initialize

    def export(self):
        """Exports stateful values to a dictionary."""
        return {
            name: element.export()
            for name, element in self._stateful_elements.items()
        }

    def _set_stateful(self, stateful_dict):
        """
        Set stateful values.

        Parameters
        ----------
        stateful_dict: dict
          Dictionary of the stateful elements and the attributes to change.

        """
        for element, attrs in stateful_dict.items():
            component = getattr(self.components, element)
            if hasattr(component, '_set_stateful'):
                component._set_stateful(attrs)
            else:
                [
                    setattr(component, attr, value)
                    for attr, value in attrs.items()
                ]

    def _build_doc(self):
        """
        Formats a table of documentation strings to help users remember
        variable names, and understand how they are translated into
        Python safe names.

        Returns
        -------
        docs_df: pandas dataframe
            Dataframe with columns for the model components:
                - Real names
                - Python safe identifiers (as used in model.components)
                - Units string
                - Documentation strings from the original model file
        """
        collector = []
        for name, pyname in self._namespace.items():
            element = getattr(self.components, pyname)
            collector.append({
                'Real Name': name,
                'Py Name': pyname,
                'Subscripts': element.subscripts,
                'Units': element.units,
                'Limits': element.limits,
                'Type': element.type,
                'Subtype': element.subtype,
                'Comment': element.__doc__.strip().strip("\n").strip()
                if element.__doc__ else None
            })

        return pd.DataFrame(
            collector
        ).sort_values(by="Real Name").reset_index(drop=True)

    def __str__(self):
        """ Return model source files """

        # JT: Might be helpful to return not only the source file, but
        # also how the instance differs from that source file. This
        # would give a more accurate view of the current model.
        string = 'Translated Model File: ' + self.py_model_file
        if hasattr(self, 'mdl_file'):
            string += '\n Original Model File: ' + self.mdl_file

        return string


class Model(Macro):
    """
    The Model class implements a stateful representation of the system.
    It inherits methods from the Macro class to integrate the model and
    access and modify model components. It also contains the main
    methods for running the model.

    The Model object will be created with components drawn from a
    translated Python model file.

    Parameters
    ----------
    py_model_file: str or pathlib.Path
        Filename of a model which has already been converted into a
        Python format.
    data_files: dict or list or str or None
        The dictionary with keys the name of file and variables to
        load the data from there. Or the list of names or name of the
        file to search the data in. Only works for TabData type object
        and it is neccessary to provide it. Default is None.
    initialize: bool
        If False, the model will not be initialize when it is loaded.
        Default is True.
    missing_values : str ("warning", "error", "ignore", "keep") (optional)
        What to do with missing values. If "warning" (default)
        shows a warning message and interpolates the values.
        If "raise" raises an error. If "ignore" interpolates
        the values without showing anything. If "keep" it will keep
        the missing values, this option may cause the integration to
        fail, but it may be used to check the quality of the data.

    See also
    --------
    :class:`pysd.py_backend.model.Macro`

    """
    def __init__(self, py_model_file, data_files, initialize, missing_values):
        """ Sets up the Python objects """
        super().__init__(py_model_file, None, None, Time(),
                         data_files=data_files)
        self.data_files = data_files
        self.missing_values = missing_values
        # set time component
        self.time.stage = 'Load'
        # set control var privately to do not change it when copying
        self.time._set_control_vars(**self.components._control_vars)
        # Attributes that are set later
        self.progress = None
        self.output = None
        self.capture_elements = None
        self.return_addresses = None
        self._stepper_mode = None
        self._submodel_tracker = {}

        if initialize:
            self.initialize()

    def initialize(self):
        """
        Initializes the simulation model.

        See also
        --------
        :func:`pysd.py_backend.model.Macro.initialize`
        :func:`pysd.py_backend.model.Model.reload`

        """
        self.time.stage = 'Initialization'
        External.missing = self.missing_values
        super().initialize()

    def run(self, params=None, return_columns=None, return_timestamps=None,
            initial_condition='original', final_time=None, time_step=None,
            saveper=None, reload=False, progress=False, flatten_output=True,
            cache_output=True, output_file=None):
        """
        Simulate the model's behavior over time.
        Return a pandas dataframe with timestamps as rows and model
        elements as columns. More detailed information and examples
        of usage are available at
        `Getting Started <https://pysd.readthedocs.io/en/master/getting_started.html#running-the-model>`__.

        Parameters
        ----------
        params: dict (optional)
            Keys are strings of model component names.
            Values are numeric or pandas Series.
            Numeric values represent constants over the model integration.
            Timeseries will be interpolated to give time-varying input.
            For more information, check the documentation of
            :func:`pysd.py_backend.model.Macro.set_components`.

        return_timestamps: list, numeric, ndarray (1D) (optional)
            Timestamps in model execution at which to return state information.
            Defaults to model-file specified timesteps.

        return_columns: list, 'step' or None (optional)
            List of string model component names, returned dataframe
            will have corresponding columns. If 'step' only variables with
            cache step will be returned. If None, variables with cache step
            and run will be returned. Default is None.

        initial_condition: str or (float, dict) (optional)
            The starting time, and the state of the system (the values of
            all the stocks) at that starting time. 'original' or 'o' uses
            model-file specified initial condition. 'current' or 'c' uses
            the state of the model after the previous execution. Other str
            objects, loads initial conditions from the pickle file with the
            given name.(float, dict) tuple lets the user specify a starting
            time (float) and (possibly partial) dictionary of initial values
            for stock (stateful) objects. Default is 'original'.
            For more information, check the documentation of
            :func:`pysd.py_backend.model.Model.set_initial_condition`

        final_time: float or None
            Final time of the simulation. If float, the given value will be
            used to compute the return_timestamps (if not given) and as a
            final time. If None the last value of return_timestamps will be
            used as a final time. Default is None.

        time_step: float or None
            Time step of the simulation. If float, the given value will be
            used to compute the return_timestamps (if not given) and
            euler time series. If None the default value from components
            will be used. Default is None.

        saveper: float or None
            Saving step of the simulation. If float, the given value will be
            used to compute the return_timestamps (if not given). If None
            the default value from components will be used. Default is None.

        reload : bool (optional)
            If True, reloads the model from the translated model file
            before making changes. Default is False.

        progress : bool (optional)
            If True, a progressbar will be shown during integration.
            Default is False.

        flatten_output: bool (optional)
            If True, once the output dataframe has been formatted will
            split the xarrays in new columns following Vensim's naming
            to make a totally flat output. Default is True.
            This argument will be ignored when passing a netCDF4 file
            path in the output_file argument.

        cache_output: bool (optional)
           If True, the number of calls of outputs variables will be increased
           in 1. This helps caching output variables if they are called only
           once. For performance reasons, if time step = saveper it is
           recommended to activate this feature, if time step << saveper
           it is recommended to deactivate it. Default is True.

        output_file: str, pathlib.Path or None (optional)
           Path of the file in which to save simulation results.
           Currently, csv, tab and nc (netCDF4) files are supported.


        Examples
        --------
        >>> model.run(params={'exogenous_constant': 42})
        >>> model.run(params={'exogenous_variable': timeseries_input})
        >>> model.run(return_timestamps=[1, 2, 3, 4, 10])
        >>> model.run(return_timestamps=10)
        >>> model.run(return_timestamps=np.linspace(1, 10, 20))
        >>> model.run(output_file="results.nc")

        See also
        --------
        :func:`pysd.py_backend.model.Macro.set_components`
        :func:`pysd.py_backend.model.Model.set_initial_condition`
        :func:`pysd.py_backend.model.Model.reload`

        """
        self._stepper_mode = False

        if reload:
            self.reload()

        self._config_simulation(params, return_columns, return_timestamps,
                                initial_condition, final_time, time_step,
                                saveper, cache_output, progress=progress)

        # instante output object
        self.output = ModelOutput(output_file)
        self.output.set_capture_elements(self.capture_elements)
        self.output.initialize(self)

        self._integrate()

        return self.output.collect(self, flatten_output)

    def set_stepper(self, output_obj, params=None, step_vars=[],
                    return_columns=None, return_timestamps=None,
                    initial_condition='original', final_time=None,
                    time_step=None, saveper=None, cache_output=True):
        """
        Configure the model stepping behavior. Examples of usage are
        available at
        `Advanced Usage <https://pysd.readthedocs.io/en/master/advanced_usage.html#running-models-one-or-more-step-s-at-a-time>`__.

        Parameters
        ----------
        output_obj: ModelOutput
            Instance of ModelOutput where the simulation results will be
            stored.

        params: dict (optional)
            Keys are strings of model component names.
            Values are numeric or pandas Series.
            Numeric values represent constants over the model integration.
            Timeseries will be interpolated to give time-varying input.

        step_vars: list
            List of variable or parameter names whose values might be
            updated after one or more simulation steps.

        return_columns: list, 'step' or None (optional)
            List of string model component names, returned dataframe
            will have corresponding columns. If 'step' only variables
            with cache step will be returned. If None, variables with
            cache step and run will be returned. Default is None.

        return_timestamps: list, numeric, ndarray (1D) (optional)
            Timestamps in model execution at which to return state
            information. Defaults to model-file specified timesteps.

        initial_condition: str or (float, dict) (optional)
            The starting time, and the state of the system (the values
            of all the stocks) at that starting time. 'original' or 'o'
            uses model-file specified initial condition. 'current' or
            'c' uses the state of the model after the previous
            execution. Other str objects, loads initial conditions from
            the pickle file with the given name.(float, dict) tuple lets
            the user specify a starting time (float) and (possibly
            partial) dictionary of initial values for stock (stateful)
            objects. Default is 'original'.

        final_time: float or None
            Final time of the simulation. If float, the given value will
            be used to compute the return_timestamps (if not given) and
            as a final time. If None the last value of return_timestamps
            will be used as a final time. Default is None.

        time_step: float or None
            Time step of the simulation. If float, the given value will
            be used to compute the return_timestamps (if not given) and
            euler time series. If None the default value from components
            will be used. Default is None.

        saveper: float or None
            Saving step of the simulation. If float, the given value will
            be used to compute the return_timestamps (if not given). If None
            the default value from components will be used. Default is None.

        cache_output: bool (optional)
           If True, the number of calls of outputs variables will be increased
           in 1. This helps caching output variables if they are called only
           once. For performance reasons, if time step = saveper it is
           recommended to activate this feature, if time step << saveper
           it is recommended to deactivate it. Default is True.

        See also
        --------
        :func:`pysd.py_backend.model.Model.step`

        """
        self.output = output_obj

        self._stepper_mode = True

        self._config_simulation(params, return_columns, return_timestamps,
                                initial_condition, final_time, time_step,
                                saveper, cache_output, step_vars=step_vars)

        self.output.set_capture_elements(self.capture_elements)
        self.output.initialize(self)
        self.output.update(self)

    def step(self, num_steps=1, step_vars={}):
        """
        Run a model step. Updates model variables first (optional), and then
        runs any number of model steps. To collect the outputs after one or
        more steps, use the collect method of the ModelOutput class.
        Examples of usage are available at
        `Advanced Usage <https://pysd.readthedocs.io/en/master/advanced_usage.html#running-models-one-or-more-step-s-at-a-time>`__

        Parameters
        ----------
        num_steps: int
            Number of steps that the iterator should run with the values
            of variables defined in step_vars argument.

        step_vars: dict
            Varibale names that should be updated before running the step
            as keys, and the actual values of the variables as values.

        Returns
        -------
        None

        See also
        --------
        :func:`pysd.py_backend.model.Model.set_stepper`

        """
        # TODO warn the user if we exceeded the final_time??
        self._set_components(step_vars, new=False)

        for _ in range(num_steps):
            self._integrate_step()
            if self.time.in_return():
                self.output.update(self)

    def _config_simulation(self, params, return_columns, return_timestamps,
                           initial_condition, final_time, time_step,
                           saveper, cache_output, **kwargs):
        """
        Internal method to set all simulation config parameters. Arguments
        to this function are those of the run and set_stepper methods.
        """
        # set control var at the beginning in case they are needed to
        # initialize any object
        self._set_control_vars(return_timestamps, final_time, time_step,
                               saveper)

        if params:
            self.set_components(params)

        if self._stepper_mode:
            for step_var in kwargs["step_vars"]:
                self._dependencies[step_var]["time"] = 1

        # update cache types after setting params
        self._assign_cache_type()

        # set initial conditions
        self.set_initial_condition(initial_condition)
        # set control vars again in case a pickle has been used
        self._set_control_vars(return_timestamps, final_time, time_step,
                               saveper)

        # progressbar only makes sense when not running step by step
        if not self._stepper_mode:
            self.progress = self._set_progressbar(kwargs["progress"])

        self.capture_elements = self._set_capture_elements(return_columns)

        # include outputs in cache if needed
        self._dependencies["OUTPUTS"] = {
            element: 1 for element in self.capture_elements["step"]
        }

        if cache_output:
            # udate the cache type taking into account the outputs
            self._assign_cache_type()

        # add constant cache to thosa variable that are constants
        self._add_constant_cache()

        # set Run mode
        self.time.stage = 'Run'

        # need to clean cache to remove the values from active_initial
        self.clean_caches()

    def _set_capture_elements(self, return_columns):
        """
        Define which variables will be stored in the output object, according
        to the return_columns passed by the user. The list is stored in the
        capture_elements attribute, which is a dictionary with keys "run" and
        "step".

        Parameters
        ----------

        return_columns:list, 'step' or None (optional)
            List of string model component names, returned dataframe
            will have corresponding columns. If 'step' only variables with
            cache step will be returned. If None, variables with cache step
            and run will be returned. Default is None.

        Returns
        -------
        capture_elements: dict
            Dictionary of list with keywords step and run.

        """
        if return_columns is None or isinstance(return_columns, str):
            return_columns = self._default_return_columns(return_columns)

        capture_elements, self.return_addresses = utils.get_return_elements(
            return_columns, self._namespace)

        # create a dictionary splitting run cached and others
        capture_elements = self._split_capture_elements(capture_elements)

        return capture_elements

    def _set_progressbar(self, progress):
        """
        Configures the progressbar, according to the user provided argument.
        If final_time or time_step are functions, then the progressbar is
        automatically disabled, regardless of the value of the progress
        argument.

        Parameters
        ----------
        progress: bool

        """
        if progress and (self.cache_type["final_time"] == "step" or
                         self.cache_type["time_step"] == "step"):
            warnings.warn(
                "The progressbar is not compatible with dynamic "
                "final time or time step. Both variables must be "
                "constants to prompt progress."
            )
            progress = False

        return progress

    def _set_control_vars(self, return_timestamps, final_time, time_step,
                          saveper):
        self.time.add_return_timestamps(return_timestamps)
        if self.time.return_timestamps is not None and not final_time:
            # if not final time given the model will end in the list
            # return timestamp (the list is reversed for popping)
            if self.time.return_timestamps:
                final_time = self.time.return_timestamps[0]
            else:
                final_time = self.time._next_return

        self.time.set_control_vars(
            final_time=final_time, time_step=time_step, saveper=saveper)

    def select_submodel(self, vars=[], modules=[], exogenous_components={},
                        inplace=True):
        """
        Select a submodel from the original model. After selecting a submodel
        only the necessary stateful objects for integrating this submodel will
        be computed. Examples of usage are available at
        `Advanced Usage <https://pysd.readthedocs.io/en/master/advanced_usage.html#selecting-and-running-a-submodel>`__.

        Parameters
        ----------
        vars: set or list of strings (optional)
            Variables to include in the new submodel.
            It can be an empty list if the submodel is only selected by
            module names. Default is an empty list.

        modules: set or list of strings (optional)
            Modules to include in the new submodel.
            It can be an empty list if the submodel is only selected by
            variable names. Default is an empty list. Can select a full
            module or a submodule by passing the path without the .py, e.g.:
            "view_1/submodule1".

        exogenous_components: dictionary of parameters (optional)
            Exogenous value to fix to the model variables that are needed
            to run the selected submodel. The exogenous_components should
            be passed as a dictionary in the same way it is done for
            set_components method. By default it is an empty dict and
            the needed exogenous components will be set to a numpy.nan value.

        inplace: bool (optional)
            If True it will modify current object and will return None.
            If False it will create a copy of the model and return it
            keeping the original model unchange. Default is True.

        Returns
        -------
        None or pysd.py_backend.model.Model
            If inplace=False it will return a modified copy of the
            original model.

        Note
        ----
        modules can be only passed when the model has been split in
        different files during translation.

        Examples
        --------
        >>> model.select_submodel(
        ...     vars=["Room Temperature", "Teacup temperature"])
        UserWarning: Selecting submodel, to run the full model again use model.reload()

        >>> model.select_submodel(
        ...     modules=["view_1", "view_2/subview_1"])
        UserWarning: Selecting submodel, to run the full model again use model.reload()
        UserWarning: Exogenous components for the following variables are necessary but not given:
            initial_value_stock1, stock3

        >>> model.select_submodel(
        ...     vars=["stock3"],
        ...     modules=["view_1", "view_2/subview_1"])
        UserWarning: Selecting submodel, to run the full model again use model.reload()
        UserWarning: Exogenous components for the following variables are necessary but not given:
            initial_value_stock1, initial_value_stock3
        Please, set them before running the model using set_components method...

        >>> model.select_submodel(
        ...     vars=["stock3"],
        ...     modules=["view_1", "view_2/subview_1"],
        ...     exogenous_components={
        ...         "initial_value_stock1": 3,
        ...         "initial_value_stock3": 5})
        UserWarning: Selecting submodel, to run the full model again use model.reload()

        See also
        --------
        :func:`pysd.py_backend.model.Model.get_vars_in_module`
        :func:`pysd.py_backend.model.Model.get_dependencies`

        """
        if inplace:
            self._select_submodel(vars, modules, exogenous_components)
        else:
            return self.copy()._select_submodel(
                vars, modules, exogenous_components)

    def _select_submodel(self, vars, modules, exogenous_components={}):
        self._submodel_tracker = {
            "vars": vars,
            "modules": modules
        }
        deps = self.get_dependencies(vars, modules)
        warnings.warn(
            "Selecting submodel, "
            "to run the full model again use model.reload()")

        # get set of all dependencies and all variables to select
        all_deps = deps.d_deps["initial"].copy()
        all_deps.update(deps.d_deps["step"])
        all_deps.update(deps.d_deps["lookup"])

        all_vars = all_deps.copy()
        all_vars.update(deps.c_vars)

        # clean dependendies and namespace dictionaries, and remove
        # the rows from the documentation
        for real_name, py_name in self._namespace.copy().items():
            if py_name not in all_vars:
                del self._namespace[real_name]
                del self._dependencies[py_name]
                self._doc.drop(
                    self._doc.index[self._doc["Real Name"] == real_name],
                    inplace=True
                )

        for py_name in self._dependencies.copy().keys():
            if py_name.startswith("_") and py_name not in deps.s_deps:
                del self._dependencies[py_name]

        # remove active initial from s_deps as they are "fake" objects
        # in dependencies
        deps.s_deps = {
            dep for dep in deps.s_deps if not dep.startswith("_active_initial")
        }

        # reassing the dictionary and lists of needed stateful objects
        self._stateful_elements = {
            name: getattr(self.components, name)
            for name in deps.s_deps
            if isinstance(getattr(self.components, name), Stateful)
        }
        self._dynamicstateful_elements = [
            getattr(self.components, name) for name in deps.s_deps
            if isinstance(getattr(self.components, name), DynamicStateful)
        ]
        self._macro_elements = [
            getattr(self.components, name) for name in deps.s_deps
            if isinstance(getattr(self.components, name), Macro)
        ]

        # keeping only needed external objects
        ext_deps = set()
        for values in self._dependencies.values():
            if "__external__" in values:
                ext_deps.add(values["__external__"])
        self._external_elements = [
            getattr(self.components, name) for name in ext_deps
            if isinstance(getattr(self.components, name), External)
        ]

        # set all exogenous values to np.nan by default
        new_components = {element: np.nan for element in all_deps}
        # update exogenous values with the user input
        [new_components.update(
            {
                utils.get_key_and_value_by_insensitive_key_or_value(
                    key,
                    self._namespace)[1]: value
            }) for key, value in exogenous_components.items()]

        self.set_components(new_components)

        # show a warning message if exogenous values are needed for a
        # dependency
        new_components = [
            key for key, value in new_components.items() if value is np.nan]
        if new_components:
            warnings.warn(
                "Exogenous components for the following variables are "
                f"necessary but not given:\n\t{', '.join(new_components)}"
                "\n\n Please, set them before running the model using "
                "set_components method...")

        # re-assign the cache_type and initialization order
        self._assign_cache_type()
        self._get_initialize_order()
        return self

    def get_dependencies(self, vars=[], modules=[]):
        """
        Get the dependencies of a set of variables or modules.

        Parameters
        ----------
        vars: set or list of strings (optional)
            Variables to get the dependencies from.
            It can be an empty list if the dependencies are computed only
            using modules. Default is an empty list.
        modules: set or list of strings (optional)
            Modules to get the dependencies from.
            It can be an empty list if the dependencies are computed only
            using variables. Default is an empty list. Can select a full
            module or a submodule by passing the path without the .py, e.g.:
            "view_1/submodule1".

        Returns
        -------
        dependencies: pysd.py_backend.utils.Dependencies
            Dependencies data object.

        Note
        ----
        modules can be only passed when the model has been split in
        different files during translation.

        Examples
        --------
        >>> print(model.get_dependencies(
        ...     vars=["Room Temperature", "Teacup temperature"]))
        Selected variables (total 1):
            room_temperature, teacup_temperature
        Stateful objects integrated with the selected variables (total 1):
            _integ_teacup_temperature

        >>> print(model.get_dependencies(
        ...     modules=["view_1", "view_2/subview_1"]))
        Selected variables (total 4):
            var1, var2, stock1, delay1
        Dependencies for initialization only (total 1):
            initial_value_stock1
        Dependencies that may change over time (total 2):
            stock3
        Stateful objects integrated with the selected variables (total 1):
            _integ_stock1, _delay_fixed_delay1

        >>> print(model.get_dependencies(
        ...     vars=["stock3"],
        ...     modules=["view_1", "view_2/subview_1"]))
        Selected variables (total 4):
            var1, var2, stock1, stock3, delay1
        Dependencies for initialization only (total 1):
            initial_value_stock1, initial_value_stock3
        Stateful objects integrated with the selected variables (total 1):
            _integ_stock1, _integ_stock3, _delay_fixed_delay1

        See also
        --------
        :func:`pysd.py_backend.model.Model.get_vars_in_module`

        """
        def check_dep(deps_obj, deps, initial=False):
            for dep in deps:
                if dep in deps_obj.c_vars or dep.startswith("__"):
                    pass
                elif dep.startswith("_"):
                    deps_obj.s_deps.add(dep)
                    dep = self._dependencies[dep]
                    check_dep(deps_obj, dep["initial"], True)
                    check_dep(deps_obj, dep["step"])
                else:
                    if initial and dep not in deps_obj.d_deps["step"]\
                       and dep not in deps_obj.d_deps["lookup"]:
                        deps_obj.d_deps["initial"].add(dep)
                    else:
                        if dep in deps_obj.d_deps["initial"]:
                            deps_obj.d_deps["initial"].remove(dep)
                        if self.get_args(dep):
                            deps_obj.d_deps["lookup"].add(dep)
                        else:
                            deps_obj.d_deps["step"].add(dep)

        dependencies = utils.Dependencies(
            {"time", "time_step", "initial_time", "final_time", "saveper"},
            {"initial": set(), "step": set(), "lookup": set()},
            set()
        )
        for var in vars:
            py_name = utils.get_key_and_value_by_insensitive_key_or_value(
                    var,
                    self._namespace)[1]
            dependencies.c_vars.add(py_name)
        for module in modules:
            dependencies.c_vars.update(self.get_vars_in_module(module))

        for var in dependencies.c_vars:
            if var == "time":
                continue
            check_dep(dependencies, self._dependencies[var])

        return dependencies

    def get_vars_in_module(self, module):
        """
        Return the name of Python vars in a module.

        Parameters
        ----------
        module: str
            Name of the module to search in.

        Returns
        -------
        vars: set
            Set of varible names in the given module.

        See also
        --------
        :func:`pysd.py_backend.model.Model.get_dependencies`

        """
        if self._modules:
            module_content = self._modules.copy()
        else:
            raise ValueError(
                "Trying to get a module from a non-modularized model")

        try:
            # get the module or the submodule content
            for submodule in module.split("/"):
                module_content = module_content[submodule]
            module_content = [module_content]
        except KeyError:
            raise NameError(
                f"Module or submodule '{submodule}' not found...\n")

        vars, new_content = set(), []

        while module_content:
            # find the vars in the module or the submodule
            for content in module_content:
                if isinstance(content, list):
                    vars.update(content)
                else:
                    [new_content.append(value) for value in content.values()]

            module_content, new_content = new_content, []

        return vars

    def copy(self, reload=False):
        """
        Create a copy of the current model.

        Parameters
        ----------
        reload: bool (optional)
            If True the model will be copied without applying to it any
            change, the copy will simply load the model again from the
            translated file. This would be equivalent to doing
            :py:func:`pysd.load` with the same arguments. Otherwise, it
            will apply the same changes that have been applied to the
            original model and update the states (faithful copy).
            Default is False.

        Warning
        -------
        The copy function will load a new model from the file and apply
        the same changes to it. If any of these changes have replaced a
        variable with a function that references other variables in the
        model, the copy will not work properly since the function will
        still reference the variables in the original model, in which
        case the function should be redefined.

        See also
        --------
        :func:`pysd.py_backend.model.Model.reload`

        """
        # initialize the new model?
        initialize = self.time.stage != 'Load'
        # create a new model
        new_model = type(self)(
            py_model_file=deepcopy(self.py_model_file),
            data_files=deepcopy(self.data_files),
            initialize=initialize,
            missing_values=deepcopy(self.missing_values)
        )
        if reload:
            # return reloaded copy
            return new_model
        # copy the values of the stateful objects
        if initialize:
            new_model._set_stateful(deepcopy(super().export()))
        # copy time object values
        new_model.time._set_time(deepcopy(self.time.export()))
        # set other components
        with warnings.catch_warnings():
            # filter warnings that have been already shown in original model
            warnings.simplefilter("ignore")
            # substract submodel
            if self._submodel_tracker:
                new_model._select_submodel(**self._submodel_tracker)
            # copy modified parameters
            new_model.set_components(self._components_setter_tracker)

        return new_model

    def reload(self):
        """
        Reloads the model from the translated model file, so that all the
        parameters are back to their original value.

        See also
        --------
        :func:`pysd.py_backend.model.Model.copy`
        :func:`pysd.py_backend.model.Model.initialize`

        """
        self.__init__(self.py_model_file, data_files=self.data_files,
                      initialize=True,
                      missing_values=self.missing_values)

    def _default_return_columns(self, which):
        """
        Return a list of the model elements tha change on time that
        does not include lookup other functions that take parameters
        or run-cached functions.

        Parameters
        ----------
        which: str or None
            If it is 'step' only cache step elements will be returned.
            Else cache 'step' and 'run' elements will be returned.
            Default is None.

        Returns
        -------
        return_columns: list
            List of columns to return

        """
        if which == 'step':
            types = ['step']
        else:
            types = ['step', 'run']

        return_columns = []

        for key, pykey in self._namespace.items():
            if pykey in self.cache_type and self.cache_type[pykey] in types\
               and not self.get_args(pykey):

                return_columns.append(key)

        return return_columns

    def _split_capture_elements(self, capture_elements):
        """
        Splits the capture elements list between those with run cache
        and others.

        Parameters
        ----------
        capture_elements: list
            Captured elements list

        Returns
        -------
        capture_dict: dict
            Dictionary of list with keywords step and run.

        """
        capture_dict = {'step': [], 'run': [], None: []}
        [capture_dict[self.cache_type[element]].append(element)
         for element in capture_elements]
        return capture_dict

    def set_initial_condition(self, initial_condition):
        """ Set the initial conditions of the integration.

        Parameters
        ----------
        initial_condition : str or (float, dict) or pathlib.Path
            The starting time, and the state of the system (the values of
            all the stocks) at that starting time. 'original' or 'o'uses
            model-file specified initial condition. 'current' or 'c' uses
            the state of the model after the previous execution. Other str
            objects, loads initial conditions from the pickle file with the
            given name.(float, dict) tuple lets the user specify a starting
            time (float) and (possibly partial) dictionary of initial values
            for stock (stateful) objects.

        Examples
        --------
        >>> model.set_initial_condition('original')
        >>> model.set_initial_condition('current')
        >>> model.set_initial_condition('exported_pickle.pic')
        >>> model.set_initial_condition((10, {'teacup_temperature': 50}))

        See also
        --------
        :func:`pysd.py_backend.model.Macro.set_initial_value`

        """
        if isinstance(initial_condition, str)\
           and initial_condition.lower() not in ["original", "o",
                                                 "current", "c"]:
            initial_condition = Path(initial_condition)

        if isinstance(initial_condition, tuple):
            self.initialize()
            self.set_initial_value(*initial_condition)
        elif isinstance(initial_condition, Path):
            self.import_pickle(initial_condition)
            self.time.set_control_vars(initial_time=self.time())
        elif isinstance(initial_condition, str):
            if initial_condition.lower() in ["original", "o"]:
                self.time.set_control_vars(
                    initial_time=self.components._control_vars["initial_time"])
                self.initialize()
        else:
            raise TypeError(
                "Invalid initial conditions. "
                + "Check documentation for valid entries or use "
                + "'help(model.set_initial_condition)'.")

    def _euler_step(self, dt):
        """
        Performs a single step in the euler integration,
        updating stateful components

        Parameters
        ----------
        dt : float
            This is the amount to increase time by this step

        """
        self.state = self.state + self.ddt() * dt

    def _integrate(self):
        """
        Performs euler integration and writes results to the out_obj.

        Returns
        -------
        None

        """

        if self.progress:
            # initialize progress bar
            progressbar = utils.ProgressBar(
                int((self.time.final_time()-self.time())/self.time.time_step())
            )
        else:
            # when None is used the update will do nothing
            progressbar = utils.ProgressBar(None)

        # performs the time stepping
        while self.time.in_bounds():
            if self.time.in_return():
                self.output.update(self)

            self._integrate_step()
            progressbar.update()

        # need to add one more time step, because we run only the state
        # updates in the previous loop and thus may be one short.
        if self.time.in_return():
            self.output.update(self)

        progressbar.finish()

    def _integrate_step(self):
        self._euler_step(self.time.time_step())
        self.time.update(self.time()+self.time.time_step())
        self.clean_caches()

    def export(self, file_name):
        """
        Export stateful values to pickle file.

        Parameters
        ----------
        file_name: str or pathlib.Path
          Name of the file to export the values.

        See also
        --------
        :func:`pysd.py_backend.model.Model.import_pickle`

        """
        warnings.warn(
            "\nCompatibility of exported states could be broken between"
            " different versions of PySD or xarray, current versions:\n"
            f"\tPySD {__version__}\n\txarray {xr.__version__}\n"
        )
        with open(file_name, 'wb') as file:
            pickle.dump(
                (self.time.export(),
                 super().export(),
                 {'pysd': __version__, 'xarray': xr.__version__}
                 ), file)

    def import_pickle(self, file_name):
        """
        Import stateful values from pickle file.

        Parameters
        ----------
        file_name: str or pathlib.Path
          Name of the file to import the values from.

        See also
        --------
        :func:`pysd.py_backend.model.Model.export_pickle`

        """
        with open(file_name, 'rb') as file:
            time_dict, stateful_dict, metadata = pickle.load(file)

        if __version__ != metadata['pysd']\
           or xr.__version__ != metadata['xarray']:  # pragma: no cover
            warnings.warn(
                "\nCompatibility of exported states could be broken between"
                " different versions of PySD or xarray. Current versions:\n"
                f"\tPySD {__version__}\n\txarray {xr.__version__}\n"
                "Loaded versions:\n"
                f"\tPySD {metadata['pysd']}\n\txarray {metadata['xarray']}\n"
                )

        self.time._set_time(time_dict)
        self._set_stateful(stateful_dict)
